package roebling;

/**
 * Title:        Roebling
 * Description:  Parametric dataflow programming.
 * Copyright:    Copyright (c) 2001
 * Company:
 * @author Simon Greenwold
 * @version 1.0
 */

import java.io.*;
import java.awt.geom.*;
import java.awt.*;

public class Enclosure extends Entity {

  public Entity[] mDirectory = new Entity[0];
  public Transformation[] mTransforms = new Transformation[0];
  public boolean mIsTopLevelStored = false;
  public AffineTransform mCompositeTrans;
  private int lastMadeTransform = -1;

  public Enclosure() {
    mCompositeTrans = new AffineTransform();
  }

  public Enclosure(String name, Enclosure enclosure) {
    super(name, enclosure);
    mCompositeTrans = new AffineTransform();
  }

  public void moveTransformationUp(Transformation trans) {
    int index = Util.find(mTransforms, trans);
    if (index < 1)
      return;
    Transformation temp = mTransforms[index - 1];
    mTransforms[index - 1] = mTransforms[index];
    mTransforms[index] = temp;
    g.scene.refresh();
  }

  public void moveTransformationDown(Transformation trans) {
    int index = Util.find(mTransforms, trans);
    if (index < 0 || index == mTransforms.length - 1)
      return;
    Transformation temp = mTransforms[index + 1];
    mTransforms[index + 1] = mTransforms[index];
    mTransforms[index] = temp;
    g.scene.refresh();
  }

  public void makeTransform() {
    if (lastMadeTransform == g.drawCycle) {
      return;
    }
    lastMadeTransform = g.drawCycle;
    mCompositeTrans.setToIdentity();
    if (mEnclosure != null) {
      mEnclosure.makeTransform();
      mCompositeTrans.concatenate(mEnclosure.mCompositeTrans);
    }
    composeTransforms();
  }

  private void composeTransforms() {
    for (int i = 0; i < mTransforms.length; i++) {
      mCompositeTrans.concatenate(mTransforms[i].mAffineTrans);
    }
  }

  public String newName(String baseName) {
    if (! contains(baseName)) {
      return baseName;
    }

    boolean found = true;
    int index = 1;
    while (found) {
      index++;
      found = contains(baseName + index);
    }
    return baseName + index;
  }

  public boolean contains(String name) {
    for (int i = 0; i < mDirectory.length; i++) {
      if (mDirectory[i].mName.equals(name))
        return true;
    }
    return false;
  }

  public boolean contains(Entity entity) {
    return Util.contains(mDirectory, entity);
  }

  public Entity getEntity(String name) {
    for (int i = 0; i < mDirectory.length; i++) {
      if (mDirectory[i].mName.equals(name))
        return mDirectory[i];
    }
    return null;
  }

  public void delete() {
    Entity[] dirCopy = (Entity[])mDirectory.clone();
    for (int i = 0; i < dirCopy.length; i++) {
      dirCopy[i].delete();
    }
    super.delete();
  }

  public void inverseTransform(Point2D point) {
    inverseTransform(point, point);
  }

  public void inverseTransform(Point2D src, Point2D dst) {
    makeTransform();
    try {
      mCompositeTrans.inverseTransform(src, dst);
    }
    catch (Exception e) {
    }
  }

  public void addTransformation(Transformation trans) {
    mTransforms = (Transformation[])Util.append(mTransforms, Types.TRANSFORMATION, trans);
  }

  public void removeTransformation(Transformation trans) {
    mTransforms = (Transformation[])Util.remove(mTransforms, Types.TRANSFORMATION, trans);
  }

  public boolean addEntity(Entity entity) {
    if (contains(entity))
        return false;
    mDirectory = (Entity[])Util.append(mDirectory, Types.ENTITY, entity);

    if (entity.isGraphicEntity())
        g.scene.addToDrawList((GraphicEntity)entity);
    return true;
  }

  public boolean removeEntity(Entity entity) {
    if (!contains(entity))
        return false;
    mDirectory = (Entity[])Util.remove(mDirectory, Types.ENTITY, entity);

    if (entity.isGraphicEntity())
        g.scene.removeFromDrawList((GraphicEntity)entity);
    return true;
  }

  public boolean isEnclosure() {
    return true;
  }

  public boolean isScene() {
    return false;
  }

  public void setFileWritingIDOffsets() {
    if (mID < g.fileWritingEntityIDOffset)
      g.fileWritingEntityIDOffset = mID;
    for (int i = 0; i < mProperties.length; i++) {
      if (mProperties[i].mID < g.fileWritingPropertyIDOffset)
        g.fileWritingPropertyIDOffset = mProperties[i].mID;
    }

    for (int i = 0; i < mDirectory.length; i++) {
      if (mDirectory[i].mID < g.fileWritingEntityIDOffset)
        g.fileWritingEntityIDOffset = mDirectory[i].mID;
      for (int j = 0; j < mDirectory[i].mProperties.length; j++) {
        if (mDirectory[i].mProperties[j].mID < g.fileWritingPropertyIDOffset)
          g.fileWritingPropertyIDOffset = mDirectory[i].mProperties[j].mID;
      }
      if (mDirectory[i].isEnclosure()) {
        ((Enclosure)mDirectory[i]).setFileWritingIDOffsets();
      }
    }

  }

  public boolean store() {
    g.fileWritingEntityIDOffset = Integer.MAX_VALUE;
    g.fileWritingPropertyIDOffset = Integer.MAX_VALUE;
    setFileWritingIDOffsets();

    File enclosureFile = new File(g.tempDirectory.getPath() + g.FS + "enclosure.xml");
    if (enclosureFile.exists()) {
      enclosureFile.delete();
      try {
        enclosureFile.createNewFile();
      }
      catch (IOException e) {
        return false;
      }
    }
    File functionDirectory = new File(g.tempDirectory + g.FS + "classes" + g.FS + "roebling");
    functionDirectory.mkdirs();
    FileUtil.clean(functionDirectory);

    XMLWriter xmlOut;
    try {
      xmlOut = new XMLWriter(new File(enclosureFile.getPath()));
      xmlOut.startDocument();
      mIsTopLevelStored = true;
      serializeEntity(xmlOut);
      mIsTopLevelStored = false;
      xmlOut.endDocument();
    }
    catch (Exception e) {
      return false;
    }
    return true;
  }

  public Entity instantiateResource(String resPath) {
    RRL rrl;
    if (g.HAS_SCENEGRAPH) {
      rrl = g.resourcesFrame.getRRL(resPath);
    }
    else {
      rrl = g.playerFrame.getRRL(resPath);
    }
    if (rrl == null)
      return null;
    return instantiateResource(rrl);
  }

  public Entity instantiateResource(RRL rrl) {
    if (rrl.isFunction) {                 // InstantiateFunction
      if (rrl.functionClass == null) {
        File classFile = FileUtil.unzip(rrl.resourceFile.getPath(), g.tempDirectory.getPath());
//          g.tr("Class name: " + classFile.getName());
        rrl.functionClass = Compiler.loadFunction(FileUtil.stripPostfix(classFile.getName()));
      }
      Function newFunc = g.scene.newFunction(rrl.functionClass,
                                             FileUtil.stripPostfix(rrl.resourceFile.getName()),
                                             this);
      newFunc.mIsInstance = true;
      newFunc.mRRL = rrl;
      return newFunc;
    }
    else if (rrl.isEnclosure) {
      if (rrl.expandedEnclosureDirectory == null) {
        rrl.expandedEnclosureDirectory = new File(g.tempDirectory.getPath() + g.FS + rrl.relativePath());
        FileUtil.unzip(rrl.resourceFile.getPath(), rrl.expandedEnclosureDirectory.getPath());
      }
      Enclosure restored = restore(rrl.expandedEnclosureDirectory);
      if (restored != null) {
        restored.mIsInstance = true;
        restored.mRRL = rrl;
        return restored;
      }
    }
    return null;
  }

  public Enclosure restore(File expandedEnclosureDirectory) {
    EnclosureParser parser = new EnclosureParser(new File(expandedEnclosureDirectory.getPath() + g.FS + "enclosure.xml"), this);
    return parser.parseFile();
  }

  protected boolean serializeEntity(XMLWriter xmlOut) {
    try {
      xmlOut.startElement("entity");

      Util.serializeElement(xmlOut, "type", "enclosure");
      if (mIsInstance) {
        Util.serializeElement(xmlOut, "instanceOf", mRRL.relativePath());
      }

      if (! super.serializeEntity(xmlOut))
        return false;

      if (!mIsInstance) {
        for (int i = 0; i < mTransforms.length; i++) {   // Transformations first because order counts
          mTransforms[i].serializeEntity(xmlOut);
        }
        for (int i = 0; i < mDirectory.length; i++) {
          if (!mDirectory[i].isTransformation())
            mDirectory[i].serializeEntity(xmlOut);
        }
      }
      xmlOut.endElement("entity");
    }
    catch (Exception e) {
        return false;
    }
    return true;
  }



}