package roebling;

/**
 * Title:        Roebling
 * Description:  Parametric dataflow programming.
 * Copyright:    Copyright (c) 2001
 * Company:
 * @author Simon Greenwold
 * @version 1.0
 */
import java.io.*;
import java.util.*;
import java.awt.geom.*;

class EnclosureParser extends XMLParser {
  private File mFile;
  private XMLReader mReader;

  private final int NONE = 0;

  private Stack mTagProcessors;
  private String mTag;
  private String mContent;

  private int TAG = 1;
  private int CONTENT = 2;
  private int mLastRead = NONE;
  private Enclosure mFirstEnclosure;
  private HashMap mLoadingPropIDToPropID = new HashMap();
  private Vector  mLoadingPropConnections = new Vector();
  private class LoadingPropConnection{
    public Property childProp;
    public Integer  parentLoadingID;

    LoadingPropConnection(Property child, Integer parent) {
      childProp = child;
      parentLoadingID = parent;
    }
  }

  public EnclosureParser(File enclosureFile, Enclosure context) {
    mReader = new XMLReader(enclosureFile, this);
    mFile = enclosureFile;

    mTagProcessors = new Stack();
    mTagProcessors.push(new EnclosureProcessor(context));
    mTag = null;
    mContent = null;
  }

  public void startElement(String tag) {
//      System.out.println("uri, localName, qName: " + uri + ", " + localName + ", " + qName);
    if (mLastRead == TAG) {
      ((TagProcessor)mTagProcessors.peek()).process(mTag, "");
    }

    mTag = tag;
    mLastRead = TAG;
  }

  public void endElement(String tag) {
    ((TagProcessor)mTagProcessors.peek()).endTag(tag);
  }
/*
  private String charsToString(char[] ch, int start, int length) {
    StringBuffer buffer = new StringBuffer();
    buffer.setLength(length);
    for (int i = 0; i < length; i++) {
      buffer.setCharAt(i, ch[i + start]);
    }
    return buffer.toString();
  }*/

  public void characters(String chars) {
    mContent = chars; //charsToString(ch, start, length);
    ((TagProcessor)mTagProcessors.peek()).process(mTag, mContent);
    mLastRead = CONTENT;
  }

  public Enclosure parseFile() {
//    try {
      mReader.parse();
      restorePropertyConnections();
      return mFirstEnclosure;
/*    }
    catch (IOException e) {
      e.printStackTrace();
    }
    catch (SAXException e) {
      e.printStackTrace();
    }
    return null;*/
  }

  private void restorePropertyConnections() {
    LoadingPropConnection curConnection;
    Property parent = null;
    Property child = null;

    for (Iterator i = mLoadingPropConnections.iterator(); i.hasNext();) {
      curConnection = (LoadingPropConnection)i.next();
      child = curConnection.childProp;
      parent = (Property)Property.mIDtoProperty.get(
                 (Integer)mLoadingPropIDToPropID.get(curConnection.parentLoadingID));

      if (parent != null && child != null) {
        Property.newRelationship(parent, child);
      }
      parent = null;
      child = null;
    }
  }

  private abstract class TagProcessor {
    abstract void process(String tag, String content);
    abstract void endTag(String tag);
  }

  private class EnclosureProcessor extends TagProcessor {
    String mType;
    String mName;
    Class  mFunctionClass;
    String mInstanceOf = "";
    Enclosure mContext;
    Entity    mCurrentEntity;
    Stack     mEnclosureStack;
    Stack     mEntityTypeStack;
    boolean   mHasNode = false;
    double    mLeft;
    double    mTop;
    double    mWidth;
    double    mHeight;
    boolean   mIsRevealed;
    ClassLoader mSystemLoader = this.getClass().getClassLoader();

    EnclosureProcessor(Enclosure context) {
      mEnclosureStack = new Stack();
      mEntityTypeStack = new Stack();
      mContext = context;
      mEnclosureStack.push(mContext);
      mCurrentEntity = context;
    }

    void process(String tag, String content) {
//      System.out.println("ENTITY Process tag: " + tag + ", " + content);
      if (tag.equals("type")) {
        mType = content;
        mFunctionClass = null;
        mEntityTypeStack.push(content);
      }
      else if (tag.equals("playerClass") && !g.HAS_SCENEGRAPH) {
        try {
          mFunctionClass = mSystemLoader.loadClass(content);
        }
        catch (Exception e) {
          mFunctionClass = null;
        }
      }
      else if (tag.equals("name")) {
        mName = mContext.newName(content);
        mCurrentEntity = makeEntity();
      }
      else if (tag.equals("instanceOf")) {
        String typeName = (String)mEntityTypeStack.pop();
        mEntityTypeStack.push(typeName + "Instance");
        if (!mType.equals("function") || mFunctionClass == null)
          mInstanceOf = content;
      }
      else if (tag.equals("property")) {
        mTagProcessors.push(new PropertyProcessor(mCurrentEntity));
      }
      else if (g.HAS_SCENEGRAPH) {
        if (tag.equals("class")) {
          if (mFunctionClass == null)
            mFunctionClass = Compiler.loadFunction(mFile.getParent() + g.FS +  //g.tempDirectory.getPath() + g.FS +
            "classes", FileUtil.getPostfix(content));
        }
        else if (tag.equals("entityNode")) {
          mCurrentEntity.makeNodeHints();
        }
        else if (tag.equals("left")) {
          mCurrentEntity.mNodeHints.left = (int)Double.parseDouble(content);
        }
        else if (tag.equals("top")) {
          mCurrentEntity.mNodeHints.top = (int)Double.parseDouble(content);
        }
        else if (tag.equals("width")) {
          mCurrentEntity.mNodeHints.width = (int)Double.parseDouble(content);
        }
        else if (tag.equals("height")) {
          mCurrentEntity.mNodeHints.height = (int)Double.parseDouble(content);
        }
        else if (tag.equals("revealed")) {
          mCurrentEntity.mNodeHints.revealed = Util.parseBoolean(content);
        }

      }
    }

    void endTag(String tag) {
      if (tag.equals("entity")) {
        if (mEntityTypeStack.pop().equals("enclosure")) {
          mEnclosureStack.pop();
          mContext = (Enclosure)mEnclosureStack.peek();
        }
      }
    }

    private Entity makeEntity() {
      Entity entity = null;
//      EntityNode node = null;
      RRL instanceRRL;
      boolean isInstance = ! mInstanceOf.equals("");

      if (mType.equals("point")) {
        entity = new Point(mName, mContext);
      }
      else if (mType.equals("line")) {
        entity = new Line(mName, mContext);
      }
      else if (mType.equals("polygon")) {
        entity = new Polygon(mName, mContext);
      }
      else if (mType.equals("text")) {
        entity = new Text(mName, mContext);
      }
      else if (mType.equals("image")) {
        entity = new Image(mName, mContext);
      }
      else if (mType.equals("button")) {
        entity = new Button(mName, mContext);
      }
      else if (mType.equals("scale")) {
        entity = new Scale(mName, mContext);
      }
      else if (mType.equals("rotation")) {
        entity = new Rotation(mName, mContext);
      }
      else if (mType.equals("translation")) {
        entity = new Translation(mName, mContext);
      }
      else if (mType.equals("function")) {
        if (isInstance) {
          if (g.HAS_SCENEGRAPH)
            instanceRRL = g.resourcesFrame.getRRL(mInstanceOf);
          else {
            instanceRRL = g.playerFrame.getRRL(mInstanceOf);
          }
          entity = mContext.instantiateResource(instanceRRL);
          entity.mRRL = instanceRRL;
          entity.mIsInstance = true;
          mInstanceOf = "";
        }
        else
          entity = g.scene.newFunction(mFunctionClass, mName, mContext);
      }
      else if (mType.equals("enclosure")) {
        if (isInstance) {
          if (g.HAS_SCENEGRAPH)
            instanceRRL = g.resourcesFrame.getRRL(mInstanceOf);
          else {
            instanceRRL = g.playerFrame.getRRL(mInstanceOf);
          }
          entity = mContext.instantiateResource(instanceRRL);
          entity.mRRL = instanceRRL;
          entity.mIsInstance = true;
          mInstanceOf = "";
        }
        else {
          entity = new Enclosure(mName, mContext);
        }
        if (mFirstEnclosure == null)
          mFirstEnclosure = (Enclosure)entity;
      }
      else
        return null;

      if (entity != null) {
        if (mType.equals("enclosure") && isInstance) {
//          node = g.sceneGraphPanel.getEntityNode(entity);
//          ((EnclosureNode)node).setIsRevealed(false);
//          node.setIsInstance(true);
        }
        else {
//          node = g.sceneGraphPanel.addNewEntityNode(entity, mContext);
        }
      }

      if (mType.equals("enclosure") && !isInstance) {
        mEnclosureStack.push(entity);
        mContext = (Enclosure)entity;
      }

      return entity;
    }
  }

  private class PropertyProcessor extends TagProcessor {
    Entity mEntity;
    Property mProperty;
    int    mValueIndex;
    PropertyType    mType;
    boolean mIsCompound;
    boolean mIsArray;
    String  mValueName = "";
    Integer mParent = null;
    Integer mLoadingID;

    Class mClass;

    PropertyProcessor(Entity entity) {
      mEntity = entity;
    }

    void process(String tag, String content) {
//      System.out.println("PROPERTY Process tag: " + tag + ", " + content);
      if (tag.equals("type")) {
        mType = g.propertyTypes[Integer.parseInt(content)];
        try {
          mProperty = (Property)mType.mClass.newInstance();
        }
        catch (Exception e) {
          return;
        }
      }
      else if (tag.equals("name")) {
        mProperty.mName = content;
      }
      else if (tag.equals("isInput")) {
        mProperty.mIsInput = Util.parseBoolean(content);
      }
      else if (tag.equals("isOutput")) {
        mProperty.mIsOutput = Util.parseBoolean(content);
      }
      else if (tag.equals("isEnclosureIn")) {
        mProperty.mIsEnclosureIn = Util.parseBoolean(content);
      }
      else if (tag.equals("isEnclosureOut")) {
        mProperty.mIsEnclosureOut = Util.parseBoolean(content);
      }
      else if (tag.equals("isArray")) {
        mIsArray = Util.parseBoolean(content);
        mProperty.isArray = mIsArray;
      }
      else if (tag.equals("isCompound")) {
        mIsCompound = Util.parseBoolean(content);
        mProperty.isCompound = mIsCompound;
      }
      else if (tag.equals("valueName")) {
        mValueName = content;
      }
      else if (tag.equals("value")) {                     // May cause trouble with empty strings?
        if (!content.equals("")) {                        // Fixed with kludgey $$$EMPTY$$$ fix
          if (content.equals("$$$EMPTY$$$"))
            mProperty.setValueByName(mValueName, "");
          else
            mProperty.setValueByName(mValueName, content);
        }
        else {
          mValueIndex = 0;
        }
      }
      else if (tag.equals("length")) {
        mProperty.setArrayLength(Integer.parseInt(content));
      }
      else if (tag.equals("index")) {
        if (content.equals("$$$EMPTY$$$"))
          mProperty.setValueByName(mValueName, mValueIndex, "");
        else
          mProperty.setValueByName(mValueName, mValueIndex, content);
        mValueIndex++;
      }

      else if (tag.equals("ID")) {
        mLoadingID = Integer.valueOf(content);
      }
      else if (tag.equals("parent")) {
        mParent = Integer.valueOf(content);
      }
      else if (g.HAS_SCENEGRAPH) {
        if (tag.equals("isFreeFloating")) {
          mProperty.setIsFreeFloating(Util.parseBoolean(content));
        }
        else if (tag.equals("freeFloatingLocationX")) {
          mProperty.mFreeFloatingLocationHint = new Point2D.Double(Double.parseDouble(content), 0.0);
        }
        else if (tag.equals("freeFloatingLocationY")) {
          mProperty.mFreeFloatingLocationHint.setLocation(mProperty.mFreeFloatingLocationHint.getX(),
                                                          Double.parseDouble(content));
        }
      }

    }
    void endTag(String tag) {
      Property actualProp;

//      System.out.println("PROPERTY END TAG: " + tag);

      if (tag.equals("property")) {
        if (mEntity.isEnclosure() && !mEntity.mIsInstance) {
          mEntity.addProperty(mProperty);
          actualProp = mProperty;
        }
        else {
          actualProp = mEntity.getProperty(mProperty.mName);
          actualProp.copyValue(mProperty);
          actualProp.valueSet();
        }
        mLoadingPropIDToPropID.put(mLoadingID, new Integer(actualProp.mID));
        if (mParent != null) {
          mLoadingPropConnections.add(new LoadingPropConnection(actualProp, mParent) );
        }

        mTagProcessors.pop();
      }
      if (tag.equals("value")) {
        mValueName = "";
      }
    }
  }
}