package roebling;

/**
 * Title:        Roebling
 * Description:  Parametric dataflow programming.
 * Copyright:    Copyright (c) 2001
 * Company:
 * @author Simon Greenwold
 * @version 1.0
 */

import java.util.*;

public class Entity {

  public String mName;
  public Enclosure mEnclosure;
  public int mID;
  public Property[] mProperties = new Property[0];

  private boolean mInUpdate;
  private boolean mInNeedsRefresh = false;
  private boolean mNeedsRefresh = true;
  public boolean mFresh = false;
  public boolean mHasOutputOnlyProperties = false;

  public boolean mIsInstance;
  public RRL     mRRL;
  public NodeHints mNodeHints = null;

  public Entity() {
/*    mID = g.nextEntityID++;
    if (g.hasSceneGraph) {
      g.sceneGraphPanel.newEntity(this);
    }*/
  }

  public Entity(String name, Enclosure enclosure) {
    mID = g.nextEntityID++;
    mName = name;
    mEnclosure = enclosure;
    mIsInstance = false;
    if (mEnclosure != null)
      enclosure.addEntity(this);
    if (g.scene != null) {
      g.scene.mEntityList = (Entity[])Util.append(g.scene.mEntityList, Types.ENTITY, this);
    }
  }

  public boolean isGraphicEntity() {
    return false;
  }

  public boolean isEnclosure() {
    return false;
  }

  public boolean isFunction() {
    return false;
  }

  public boolean isTransformation() {
    return false;
  }

  public void addProperty(Property property) {
    addProperty(property, mProperties.length);
  }

  public void addProperty(Property property, int index) {
    mProperties = (Property[])Util.insert(mProperties, Types.PROPERTY, property, index);
    property.mOwner = this;
    if (property.isOutput() && !property.isInput())
      mHasOutputOnlyProperties = true;
  }

  public void removeProperty(Property property) {
    mProperties = (Property[])Util.remove(mProperties, Types.PROPERTY, property);
  }

  public String getNewPropertyName(String baseName) {
    boolean found = true;
    int index = 0;
    while (found) {
      index++;
      found = (getProperty(baseName + index) != null);
    }
    return baseName + index;
  }

  public Property getProperty(String propertyName) {
    for (int i = 0; i < mProperties.length; i++) {
      if (mProperties[i].mName.equals(propertyName))
        return mProperties[i];
    }
    return null;
  }

  protected void readInputs() {}
  protected void writeOutputs() {}
  public void run() {}

  public void update() {
    if (mInUpdate)
      return;
    mInUpdate = true;
    if (! mFresh) {
//            System.out.println(getName() + " updating.");

      readInputs();
      mFresh = true;
      run();
      writeOutputs();
    }
    mInUpdate = false;
  }

  public void delete() {
    if (mEnclosure == null)   // don't delete the scene
      return;

    for (int i = 0; i < mProperties.length; i++) {
        mProperties[i].clearConnections();
    }
    mEnclosure.removeEntity(this);
    g.scene.mEntityList = (Entity[])Util.remove(g.scene.mEntityList, Types.ENTITY, this);
  }

  public void needsRefresh() {
    if (! mFresh && mHasOutputOnlyProperties) {
      return;
    }

    if (mInNeedsRefresh)
      return;
    mInNeedsRefresh = true;

    mFresh = false;
    for (int i = 0; i < mProperties.length; i++) {
      if (mProperties[i].mIsOutput) {
        mProperties[i].markChildrenForRefresh();
      }
    }
    mInNeedsRefresh = false;
  }

  // Instance-related calls
  public void reloadFromResource() {
  }

  protected boolean serializeEntity(XMLWriter xmlOut) {
    try {
      Util.serializeElement(xmlOut, "name", mName);
      Util.serializeElement(xmlOut, "ID", mID - g.fileWritingEntityIDOffset);

      serializeProperties(xmlOut);

      if (g.HAS_SCENEGRAPH) {     // Compiler switch (code will not appear in player)
        EntityNode node = g.sceneGraphPanel.getEntityNode(this);
        if (node != null) {
          node.serializeNode(xmlOut);
        }
        else if (mNodeHints != null) {
          serializeNodeHints(xmlOut);
        }
      }

    }
    catch (Exception e) {
      return false;
    }
    return true;
  }

  protected boolean serializeProperties(XMLWriter xmlOut) {
    try {
      for (int i = 0; i < mProperties.length; i++) {
        mProperties[i].serializeProperty(xmlOut);
      }
    }
    catch (Exception e) {
      return false;
    }
    return true;
  }

  private boolean serializeNodeHints(XMLWriter xmlOut) {
    try {
      xmlOut.startElement("entityNode");
      Util.serializeElement(xmlOut, "left", mNodeHints.left);
      Util.serializeElement(xmlOut, "top", mNodeHints.top);
      Util.serializeElement(xmlOut, "width", mNodeHints.width);
      Util.serializeElement(xmlOut, "height", mNodeHints.height);
      if (isEnclosure()) {
        Util.serializeElement(xmlOut, "revealed", mNodeHints.revealed);
      }
//      serializeProperties(xmlOut);      // Don't serialize prop nodes for now
      xmlOut.endElement("entityNode");
    }
    catch (Exception e) {
      return false;
    }
    return true;
  }

  public void makeNodeHints() {
    mNodeHints = new NodeHints();
  }

  public class NodeHints {       // These store info about the graph node before such a node is created
    public int top;
    public int left;
    public int width;
    public int height;
    public boolean revealed;
  }
}