package roebling;

import javax.swing.JPanel;

/**
 * Title:        ParaMour
 * Description:  Parametric design tool.
 * Copyright:    Copyright (c) 2001
 * Company:      NewBricks
 * @author Simon Greenwold
 * @version 1.0
 */

import java.awt.Color;
import java.awt.*;
import java.awt.geom.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import java.io.*;

public class EntityNode extends GraphNode {

    protected static final int TITLE_BAR_HEIGHT = 16;
    protected static final int CONTROL_BAR_HEIGHT = 12;
    public static final int X_INSET = 3;
    public static final int Y_INSET = 1;
    protected static final int BORDER_THICKNESS = 4;
    protected static final int MIN_PROPERTY_SEPARATION = 8;
    public static final int PROPERTY_LABEL_X_OFFSET = 16;
    public static final int PROPERTY_LABEL_Y_OFFSET = -2;
    protected static final int DEFAULT_HEIGHT = 140;
    protected static final int CORNER_DRAG_OFFSET = 12;
    protected static final int TITLE_X_OFFSET = 28;
    protected static final int MINIMUM_WIDTH = 100;

    protected int mDragMode = NONE;
    private int mDragStartWidth;
    private int mDragStartHeight;
    private int mDragStartX;
    private int mDragStartY;

    public PropertyEditPanel mPropertyEditPanel;
    public boolean mEditingProperty;
    protected NameEdit mNameEdit;
    protected JLabel mInstanceOfLabel;

    protected JTextField mDummyField;
    protected Component mLastFocus = null;

    public static final int PROPERTY_FUNCTION_SIDE_AMOUNT = 30;

    // Drag modes
    protected static final int NONE = 0;
    protected static final int TOP_LEFT_CORNER = 1;
    protected static final int TOP_RIGHT_CORNER = 2;
    protected static final int BOTTOM_LEFT_CORNER = 3;
    protected static final int BOTTOM_RIGHT_CORNER = 4;
    protected static final int LEFT_SIDE = 5;
    protected static final int RIGHT_SIDE = 6;
    protected static final int TOP_SIDE = 7;
    protected static final int BOTTOM_SIDE = 8;

    protected RoundRectangle2D mOuterBounds = new RoundRectangle2D.Double();
    protected RoundRectangle2D mInnerBounds = new RoundRectangle2D.Double();

    private static final Color GRAPHIC_ENTITY_COLOR = new Color(204, 255, 0, 150);
    private static BasicStroke mRectStroke = new BasicStroke(1.0f,
                                                  BasicStroke.CAP_BUTT,
                                                  BasicStroke.JOIN_MITER);

    protected static final int PROPERTY_FONT_SIZE = 10;
    protected static Font mPropertyFont = new Font("sansserif", Font.PLAIN, PROPERTY_FONT_SIZE);

    protected java.awt.Point mMouseDownOffset = new java.awt.Point();
    protected boolean mIsSelected = false;
    protected Area mSelectedBorder = null;

    protected Entity mEntity;

    protected PropertyNode[] mPropertyNodes;
    private Integer[] mPropertyYLocations = new Integer[0];
    protected EntityNode mSelf;

    protected ImageIcon mEntityImage;
    protected JLabel mName;

    public EntityNode() {
        super();
        mSelf = this;
        mPropertyNodes = new PropertyNode[0];
        try {
            jbInit();
        }
        catch(Exception e) {
            e.printStackTrace();
        }
        setOpaque(false);

        addComponentListener(new ComponentAdapter() {
            public void componentResized(ComponentEvent e) {
                resized();
            }
        });

        addMouseListener(new MouseAdapter() {
            public void mousePressed(MouseEvent e) {
                if (g.sceneGraphPanel.getTool() == SceneGraphPanel.ARROW_TOOL) {
                    mMouseDownOffset.setLocation(e.getX(), e.getY());
                    mDragStartWidth = getWidth();
                    mDragStartHeight = getHeight();
                    mDragStartX = (int)getLocation().getX();
                    mDragStartY = (int)getLocation().getY();
                    if (g.sceneGraphPanel != null && !mDidDelete ) {
                        g.sceneGraphPanel.setSelected(mSelf);
                    }
                    if (mDidDelete) {
                      mDidDelete = false;
                    }
                    parseClick(e);
                }
            }
            public void mouseReleased(MouseEvent e) {
                mDragMode = NONE;
                Point2D.Double releasePoint = new Point2D.Double();
                if (g.sceneGraphPanel.getTool() == SceneGraphPanel.ARROW_TOOL) {
                    repaint();
                }
                else if (g.sceneGraphPanel.getTool() >= SceneGraphPanel.NUMBER_PROPERTY_TOOL &&
                         g.sceneGraphPanel.getTool() <= SceneGraphPanel.CUSTOM_PROPERTY_TOOL) {
                    releasePoint.setLocation(e.getX(), e.getY());
                    makeNewProperty(releasePoint, g.sceneGraphPanel.mPropertyToolType);
                    g.sceneGraphPanel.setTool(SceneGraphPanel.ARROW_TOOL);
                }
            }
        });
        addMouseMotionListener(new MouseMotionAdapter() {
            public void mouseDragged(MouseEvent e) {
                if (g.sceneGraphPanel.getTool() == SceneGraphPanel.ARROW_TOOL)
                    drag(e);
            }
        });

        addComponentListener(new ComponentAdapter() {
            public void componentMoved(ComponentEvent e) {
                for (int i = 0; i < mPropertyNodes.length; i++) {
                    mPropertyNodes[i].moveConnections();
                }
            }
        });
    }

    //User clicked a new property onto the scene
    protected void makeNewProperty(Point2D releasePoint, int propertyType) {

        Property newProp;
        String propName = "";
//        Point2D localPoint = new Point2D.Double();
//        localPoint.setLocation(releasePoint);
        if (mEntity.isGraphicEntity())    // Don't add properties to graphic entities
          return;

        propName = mEntity.getNewPropertyName(g.propertyTypes[propertyType].mName);
        newProp = Property.newProperty(propName, propertyType);

        int width = getWidth();

        int yIndex = 0;
        for (yIndex = 0; yIndex < mPropertyYLocations.length; yIndex++) {
          if (mPropertyYLocations[yIndex].doubleValue() > releasePoint.getY())
            break;
        }

        if (isEnclosureNode()) {
            newProp.setIsOutput(true);
            newProp.setIsInput(true);

            if (Util.near(0.0, releasePoint.getX(), PROPERTY_FUNCTION_SIDE_AMOUNT)) {
                newProp.setIsEnclosureIn(true);
            }
            else if (Util.near(width, releasePoint.getX(), PROPERTY_FUNCTION_SIDE_AMOUNT)) {
                newProp.setIsEnclosureOut(true);
            }
            else {
              newProp.setIsFreeFloating(true);
            }

/*
            if (Util.near(0.0, releasePoint.getX(), width / 2)) {
                newProp.setIsEnclosureIn(true);
            }
            else {
                newProp.setIsEnclosureOut(true);
            }
*/
        }
        else {
            if (Util.near(0.0, releasePoint.getX(), PROPERTY_FUNCTION_SIDE_AMOUNT)) {
                newProp.setIsOutput(false);
            }
            else if (Util.near(width, releasePoint.getX(), PROPERTY_FUNCTION_SIDE_AMOUNT)) {
                newProp.setIsInput(false);
            }
        }
        newProp.isArray = g.sceneGraphFrame.mArrayProps;
        mEntity.addProperty(newProp);

        PropertyNode newPropNode;
        if (newProp.isFreeFloating())
          newPropNode = ((EnclosureNode)this).addFreeFloatingProperty(newProp, releasePoint);
        else
          newPropNode = addProperty(newProp, Math.max(0, yIndex));
        editPropNode(newPropNode);
    }

    public void delete() {
        mEnclosureNode.removeGraphNode(this);
        g.sceneGraphPanel.setSelected(mEnclosureNode);
        mEntity.delete();
        g.scene.refresh();
    }

    protected int getTitleBarHeight() {
        return TITLE_BAR_HEIGHT;
    }

    protected void loadEntityImage() {
        if (mEntity == null)
            return;
        else if (mEntity.isFunction()) {
            mEntityImage = new ImageIcon("images/smallFunction.gif");
        }
        else if (mEntity.isEnclosure()) {
            mEntityImage = new ImageIcon("images/smallEnclosure.gif");
        }
        else if (Types.BUTTON.isInstance(mEntity)) {
            mEntityImage = new ImageIcon("images/smallButton.gif");
        }
        else if (Types.TEXT.isInstance(mEntity)) {
            mEntityImage = new ImageIcon("images/smallText.gif");
        }
        else if (Types.POINT.isInstance(mEntity)) {
            mEntityImage = new ImageIcon("images/smallPoint.gif");
        }
        else if (Types.LINE.isInstance(mEntity)) {
            mEntityImage = new ImageIcon("images/smallLine.gif");
        }
        else if (Types.POLYGON.isInstance(mEntity)) {
            mEntityImage = new ImageIcon("images/smallPoly.gif");
        }
        else if (Types.IMAGE.isInstance(mEntity)) {
            mEntityImage = new ImageIcon("images/smallImage.gif");
        }
      }

    public void setSize(int width, int height) {
      mDummyField.requestFocus();
      super.setSize(width, height);
    }

    public void setLocation(int x, int y) {
      mDummyField.requestFocus();
      super.setLocation(x, y);
      if (mEditingProperty) {
        if (mPropertyEditPanel.mLastFocus != null) {
          mPropertyEditPanel.mLastFocus.requestFocus();
        }
      }
      else if (mLastFocus != null) {
        mLastFocus.requestFocus();
      }
    }

    public void resized() {
        mOuterBounds.setRoundRect(X_INSET, Y_INSET, getWidth() - 2 * X_INSET, getHeight() - 2 * Y_INSET, 10, 10);
        mInnerBounds.setRoundRect(X_INSET + BORDER_THICKNESS, Y_INSET + BORDER_THICKNESS, getWidth() - 2 * (X_INSET + BORDER_THICKNESS), getHeight() - 2 * (Y_INSET + BORDER_THICKNESS), 8, 8);
        moveProperties();
        if (mIsSelected)
            calculateSelectedBorder();
        mName.setSize(getWidth() - (2 * (X_INSET + BORDER_THICKNESS) + TITLE_X_OFFSET), 16);
        mNameEdit.setSize(getWidth() - (2 * (X_INSET + BORDER_THICKNESS) + TITLE_X_OFFSET), 16);
        if (mPropertyEditPanel != null)
          mPropertyEditPanel.resized();
        mInstanceOfLabel.setBounds(12, getHeight() - 36, getWidth() - 30, 24);
        if (mLastFocus != null)
          mLastFocus.requestFocus();
    }

    public void moveProperties() {
        if (mPropertyYLocations == null || mPropertyYLocations.length == 0) {
          if (mPropertyNodes != null && mPropertyNodes.length > 0) {
            for (int i = 0;  i < mPropertyNodes.length; i++) {
              if (mPropertyNodes[i].mProperty.isFreeFloating()) {
                mPropertyNodes[i].positionLabel();
              }
            }
          }
          return;
        }

        int leftAlign;

        int dy =  (getHeight() - (getTitleBarHeight() + CONTROL_BAR_HEIGHT + 2 * BORDER_THICKNESS)) / mPropertyYLocations.length;
        int yPos = getTitleBarHeight() + (dy / 2);
        for (int i = 0; i < mPropertyYLocations.length; i++) {
            mPropertyYLocations[i] = new Integer(yPos + i * dy);
        }

        for (int i = 0;  i < mPropertyNodes.length; i++) {
            if (mPropertyNodes[i].mProperty.isFreeFloating()) {
              mPropertyNodes[i].positionLabel();
              continue;
            }

            if (mPropertyNodes[i].isOnLeft()) {
                mPropertyNodes[i].setLocation(0, mPropertyYLocations[mPropertyNodes[i].getYLocationOffset()].intValue());
            }
            else {
                mPropertyNodes[i].setLocation(getWidth() - mPropertyNodes[i].getWidth(), mPropertyYLocations[mPropertyNodes[i].getYLocationOffset()].intValue());
            }
            mPropertyNodes[i].moveConnections();
            mPropertyNodes[i].positionLabel();
        }

    }

    protected void drag(MouseEvent e) {
        if (mDragMode == NONE) {
            setLocation((int)(getLocation().getX() + e.getX() - mMouseDownOffset.getX()), (int)(getLocation().getY() + e.getY() - mMouseDownOffset.getY()));
            return;
        }

        int dx = (int)(e.getX() - mMouseDownOffset.getX());
        int dy = (int)(e.getY() - mMouseDownOffset.getY());
        int x;
        int y;
        int width;
        int height;

        if (mDragMode == BOTTOM_RIGHT_CORNER) {
            if (mDragStartWidth + dx < getMinimumWidth())
                width = getMinimumWidth();
            else
                width = mDragStartWidth + dx;
            if (mDragStartHeight + dy < getMinimumHeight())
                height = getMinimumHeight();
            else
                height = mDragStartHeight + dy;

            setSize(width, height);
        }
        if (mDragMode == RIGHT_SIDE) {
            if (mDragStartWidth + dx < getMinimumWidth())
                width = getMinimumWidth();
            else
                width = mDragStartWidth + dx;
            setSize(width, mDragStartHeight);
        }
        if (mDragMode == LEFT_SIDE) {
            if (getWidth() - dx < getMinimumWidth()) {
                width = getMinimumWidth();
                x = (int)getLocation().getX();
            }
            else {
                width = getWidth() - dx;
                x = (int)getLocation().getX() + dx;
            }
            reshape(x, mDragStartY, width, mDragStartHeight);
        }
        if (mDragMode == BOTTOM_SIDE) {
            if (mDragStartHeight + dy < getMinimumHeight())
                height = getMinimumHeight();
            else
                height = mDragStartHeight + dy;

            setSize(mDragStartWidth, height);
        }
        if (mDragMode == TOP_SIDE) {
            if (getHeight() - dy < getMinimumHeight()) {
                height = getMinimumHeight();
                y = (int)getLocation().getY();
            }
            else {
                height = getHeight() - dy;
                y = (int)getLocation().getY() + dy;
            }

            reshape(mDragStartX, y, mDragStartWidth, height);
        }
        if (mDragMode == TOP_LEFT_CORNER) {
            if (getWidth() - dx < getMinimumWidth()) {
                width = getMinimumWidth();
                x = (int)getLocation().getX();
            }
            else {
                width = getWidth() - dx;
                x = (int)getLocation().getX() + dx;
            }
            if (getHeight() - dy < getMinimumHeight()) {
                height = getMinimumHeight();
                y = (int)getLocation().getY();
            }
            else {
                height = getHeight() - dy;
                y = (int)getLocation().getY() + dy;
            }

            reshape(x, y, width, height);
        }
        if (mDragMode == TOP_RIGHT_CORNER) {
            if (getHeight() - dy < getMinimumHeight()) {
                height = getMinimumHeight();
                y = (int)getLocation().getY();
            }
            else {
                height = getHeight() - dy;
                y = (int)getLocation().getY() + dy;
            }
            if (mDragStartWidth + dx < getMinimumWidth())
                width = getMinimumWidth();
            else
                width = mDragStartWidth + dx;

            reshape(mDragStartX, y, width, height);
        }
        if (mDragMode == BOTTOM_LEFT_CORNER) {
            if (mDragStartHeight + dy < getMinimumHeight())
                height = getMinimumHeight();
            else
                height = mDragStartHeight + dy;
            if (getWidth() - dx < getMinimumWidth()) {
                width = getMinimumWidth();
                x = (int)getLocation().getX();
            }
            else {
                width = getWidth() - dx;
                x = (int)getLocation().getX() + dx;
            }

            reshape(x, mDragStartY, width, height);
        }
        resized();
    }

    public void setSelected(boolean selected) {
        mIsSelected = selected;
        if (mIsSelected)
            calculateSelectedBorder();
        repaint();
    }

    private void calculateSelectedBorder() {
        mSelectedBorder = new Area(mOuterBounds);
        mSelectedBorder.subtract(new Area(mInnerBounds));
    }

    public boolean contains(int x, int y) {
        if (mOuterBounds.contains(x, y))
            return true;

        int localX;
        int localY;
        for (int i = 0; i < mPropertyNodes.length; i++) {
            localX = x - (int)mPropertyNodes[i].getLocation().getX();
            localY = y - (int)mPropertyNodes[i].getLocation().getY();
            if (mPropertyNodes[i].contains(localX, localY)) {
                return true;
            }
        }
        return false;
    }

    public boolean contains(java.awt.Point p) {
        return contains((int)p.getX(), (int)p.getY());
    }

    public Color getFillColor() {
        return GRAPHIC_ENTITY_COLOR;
    }

    public void paintComponent(Graphics g) {
        super.paintComponent(g);
        Graphics2D g2 = (Graphics2D)g;
        g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);

        g2.setStroke(mRectStroke);
        if (mIsSelected) {
            g2.setPaint(Color.red);
            g2.fill(mSelectedBorder);
            g2.setPaint(getFillColor());
            g2.fill(mInnerBounds);
        }
        else {
            g2.setPaint(getFillColor());
            g2.fill(mOuterBounds);
        }

        g2.setPaint(Color.black);
        g2.draw(mOuterBounds);
        g2.draw(mInnerBounds);

        // Line to property edit
        if (mEditingProperty) {
          int y = (int)mPropertyEditPanel.mPropertyNode.getLocation().getY();
          int x = (int)mPropertyEditPanel.mPropertyNode.getLocation().getX();
          if (mPropertyEditPanel.mPropertyNode.isInput())
            g2.drawLine(x+6, y + 5, x + 20, y + 5);
          else
            g2.drawLine(x-14, y + 5, x, y + 5);
          g2.setStroke(mRectStroke);
        }

        //Corner offset lines
        g2.drawLine(X_INSET, Y_INSET + CORNER_DRAG_OFFSET, X_INSET + BORDER_THICKNESS, Y_INSET + CORNER_DRAG_OFFSET);
        g2.drawLine(X_INSET, getHeight() - (Y_INSET + CORNER_DRAG_OFFSET), X_INSET + BORDER_THICKNESS, getHeight() - (Y_INSET + CORNER_DRAG_OFFSET));
        g2.drawLine(getWidth() - X_INSET, Y_INSET + CORNER_DRAG_OFFSET, getWidth() - (X_INSET + BORDER_THICKNESS), Y_INSET + CORNER_DRAG_OFFSET);
        g2.drawLine(getWidth() - X_INSET, getHeight() - (Y_INSET + CORNER_DRAG_OFFSET), getWidth() - (X_INSET + BORDER_THICKNESS), getHeight() - (Y_INSET + CORNER_DRAG_OFFSET));

        g2.drawLine(X_INSET + CORNER_DRAG_OFFSET, Y_INSET, X_INSET + CORNER_DRAG_OFFSET, Y_INSET + BORDER_THICKNESS);
        g2.drawLine(X_INSET + CORNER_DRAG_OFFSET, getHeight() - Y_INSET, X_INSET + CORNER_DRAG_OFFSET, getHeight() - (Y_INSET + BORDER_THICKNESS));
        g2.drawLine(getWidth() - (X_INSET + CORNER_DRAG_OFFSET), Y_INSET, getWidth() - (X_INSET + CORNER_DRAG_OFFSET), Y_INSET + BORDER_THICKNESS);
        g2.drawLine(getWidth() - (X_INSET + CORNER_DRAG_OFFSET), getHeight() - Y_INSET, getWidth() - (X_INSET + CORNER_DRAG_OFFSET), getHeight() - (Y_INSET + BORDER_THICKNESS));

        if (mEntityImage != null)
            mEntityImage.paintIcon(this, g2, X_INSET + BORDER_THICKNESS + 2, Y_INSET + BORDER_THICKNESS + 1);

        g2.drawLine(X_INSET + BORDER_THICKNESS, Y_INSET + BORDER_THICKNESS + getTitleBarHeight(),
                    getWidth() - (X_INSET + BORDER_THICKNESS), Y_INSET + BORDER_THICKNESS + getTitleBarHeight());
        g2.drawLine(X_INSET + BORDER_THICKNESS, getHeight() - (Y_INSET + BORDER_THICKNESS + CONTROL_BAR_HEIGHT),
                    getWidth() - (X_INSET + BORDER_THICKNESS), getHeight() - (Y_INSET + BORDER_THICKNESS + CONTROL_BAR_HEIGHT));
        g2.drawLine(X_INSET + BORDER_THICKNESS + 20, Y_INSET + BORDER_THICKNESS,
                    X_INSET + BORDER_THICKNESS + 20, Y_INSET + BORDER_THICKNESS + getTitleBarHeight());


    }

    public void setEntity(Entity entity) {
        mEntity = entity;
        //resetPropertyNodes();
        setSize(getMinimumWidth(), getMinimumHeight());
        addPropertyNodes();

        if (mEntityImage == null) {
            loadEntityImage();
        }
        mName.setText(mEntity.mName);
        resized();
    }

    public void setName(String name) {
      mName.setText(name);
    }

/*    public getPropertyNode(String name, boolean inputSide) {
        PropertyNode propNode;
        for (int i = 0; i < mPropertyNodes.length; i++) {
            propNode = mPropertyNodes[i];
            if (propNode.getProperty().getName().equals()
        }
    }*/

    protected void addPropertyNodes() {
      Property curProp;
      for (int i = 0; i < mEntity.mProperties.length; i++) {
        curProp = mEntity.mProperties[i];
        addProperty(curProp, i);
      }
    }

  public PropertyNode addProperty(Property property, int listOffset) {
      PropertyNode newPropNode;

      if (mPropertyNodes == null)
          mPropertyNodes = new PropertyNode[0];

      for (int i = 0; i < mPropertyNodes.length; i++) {
          if (mPropertyNodes[i].getYLocationOffset() >= listOffset)
              mPropertyNodes[i].setYLocationOffset(mPropertyNodes[i].getYLocationOffset() + 1);
      }
      newPropNode = placePropertyNode(property, listOffset);
      moveProperties();
      resized();
      return newPropNode;
  }

  protected void addPropertyConnections(EnclosureNode enclosureNode) {
    Property props[] = mEntity.mProperties;
    Property prop;
    PropertyNode parentNode;
    PropertyNode childNode;

    for (int i = 0; i < props.length; i++) {
      prop = props[i];
      if (prop.mParent != null) {
        parentNode = g.sceneGraphPanel.getOutputNode(prop.mParent);
        childNode = g.sceneGraphPanel.getInputNode(prop);
        if (parentNode != null && childNode != null)
          enclosureNode.newConnection(parentNode, childNode);
      }


    }
  }

  public void removeProperty(Property property) {
    if (mPropertyNodes == null) {
      mPropertyNodes = new PropertyNode[0];
      return;
    }

    boolean removedAny = true;
    while (removedAny) {
      removedAny = false;
      for (int i = 0; i < mPropertyNodes.length; i++) {
        if (mPropertyNodes[i].getProperty() == property) {
          removedAny = true;
          remove(mPropertyNodes[i]);
          if (mPropertyNodes[i].getLabel() != null)
            remove(mPropertyNodes[i].getLabel());
          mPropertyNodes = (PropertyNode[])Util.remove(mPropertyNodes, Types.PROPERTY_NODE, i);
          break;
        }
      }
    }
    Property[] properties = property.mOwner.mProperties;
    mPropertyYLocations = new Integer[properties.length];
    for (int i = 0; i < mPropertyNodes.length; i++) {
      mPropertyNodes[i].setYLocationOffset(Util.find(properties, mPropertyNodes[i].getProperty()));
    }
    moveProperties();
    resized();
  }

    // destroys all property connections to & from nodes
    public void resetPropertyNodes() {
        if (mPropertyNodes != null) {
            for (int i = 0; i < mPropertyNodes.length; i++) {
                remove(mPropertyNodes[i]);
                if (mPropertyNodes[i].getLabel() != null) {
                    remove(mPropertyNodes[i].getLabel());
                }
            }
        }
        mPropertyNodes = new PropertyNode[0];
        mPropertyYLocations = new Integer[0];

        Property[] properties = mEntity.mProperties;
        if (properties.length == 0)
            return;

        int yLocationOffset = 0;
        for (int i = 0; i < properties.length; i++) {
            placePropertyNode(properties[i], yLocationOffset);
            yLocationOffset++;
        }
        moveProperties();
    }

  private PropertyNode placePropertyNode(Property property, int yLocationOffset) {
    PropertyNode newProp = null;
    JLabelNoAntialias newLabel;

    if (   !(property.isEnclosureIn() || property.isEnclosureOut()) &&
            (property.isInput() && property.isOutput())) {
      newLabel = new JLabelNoAntialias(property.mName);
      newProp = new PropertyNode(property, this, newLabel);
      //property.setInputPropertyNode(newProp);
      g.sceneGraphPanel.setInputPropertyNode(property, newProp);
      mPropertyNodes = (PropertyNode[])Util.append(mPropertyNodes, Types.PROPERTY_NODE, newProp);
      mPropertyYLocations = (Integer[])Util.append(mPropertyYLocations, new Integer(0).getClass(), new Integer(0));//new Integer(yPos));
      newProp.setYLocationOffset(yLocationOffset);

      add(newLabel);
      add(newProp);
      newProp.setIsInput(true);

      newLabel.setFont(mPropertyFont);
      newLabel.setForeground(Color.black);

      newProp = new PropertyNode(property, this, null);
      g.sceneGraphPanel.setOutputPropertyNode(property, newProp);
      mPropertyNodes = (PropertyNode[])Util.append(mPropertyNodes, Types.PROPERTY_NODE, newProp);
      newProp.setIsOutput(true);
      newProp.setYLocationOffset(yLocationOffset);

      add(newProp);
    }
    else if (property.isEnclosureIn()) {
      newLabel = new JLabelNoAntialias(property.mName);
      newProp = new PropertyNode(property, this, newLabel);
      //property.setInputPropertyNode(newProp);
      //property.setOutputPropertyNode(newProp);
      g.sceneGraphPanel.setInputPropertyNode(property, newProp);
      g.sceneGraphPanel.setOutputPropertyNode(property, newProp);
      mPropertyNodes = (PropertyNode[])Util.append(mPropertyNodes, Types.PROPERTY_NODE, newProp);
      mPropertyYLocations = (Integer[])Util.append(mPropertyYLocations, new Integer(0).getClass(), new Integer(0));//new Integer(yPos));
      newProp.setYLocationOffset(yLocationOffset);

      add(newLabel);
      add(newProp);
      newProp.setIsInput(true);
      newProp.setIsOutput(true);
      newProp.setIsEnclosureIn(true);

      newLabel.setFont(mPropertyFont);
      newLabel.setForeground(Color.black);
    }
    else if (property.isEnclosureOut()) {
      newLabel = new JLabelNoAntialias(property.mName);
      newProp = new PropertyNode(property, this, newLabel);
      //property.setInputPropertyNode(newProp);
      //property.setOutputPropertyNode(newProp);
      g.sceneGraphPanel.setInputPropertyNode(property, newProp);
      g.sceneGraphPanel.setOutputPropertyNode(property, newProp);
      mPropertyNodes = (PropertyNode[])Util.append(mPropertyNodes, Types.PROPERTY_NODE, newProp);
      mPropertyYLocations = (Integer[])Util.append(mPropertyYLocations, new Integer(0).getClass(), new Integer(0));//new Integer(yPos));
      newProp.setYLocationOffset(yLocationOffset);

      add(newLabel);
      add(newProp);
      newProp.setIsOutput(true);
      newProp.setIsInput(true);
      newProp.setIsEnclosureOut(true);

      newLabel.setFont(mPropertyFont);
      newLabel.setForeground(Color.black);
      newLabel.setHorizontalAlignment(JLabel.RIGHT);
    }
    else if (property.isInput()) {
      newLabel = new JLabelNoAntialias(property.mName);
      newProp = new PropertyNode(property, this, newLabel);
      //property.setInputPropertyNode(newProp);
      g.sceneGraphPanel.setInputPropertyNode(property, newProp);
      mPropertyNodes = (PropertyNode[])Util.append(mPropertyNodes, Types.PROPERTY_NODE, newProp);
      mPropertyYLocations = (Integer[])Util.append(mPropertyYLocations, new Integer(0).getClass(), new Integer(0));//new Integer(yPos));
      newProp.setYLocationOffset(yLocationOffset);

      add(newLabel);
      add(newProp);
      newProp.setIsInput(true);

      newLabel.setFont(mPropertyFont);
      newLabel.setForeground(Color.black);
    }
    else if (property.isOutput()) {
      newLabel = new JLabelNoAntialias(property.mName);
      newProp = new PropertyNode(property, this, newLabel);
      //property.setOutputPropertyNode(newProp);
      g.sceneGraphPanel.setOutputPropertyNode(property, newProp);
      mPropertyNodes = (PropertyNode[])Util.append(mPropertyNodes, Types.PROPERTY_NODE, newProp);
      mPropertyYLocations = (Integer[])Util.append(mPropertyYLocations, new Integer(0).getClass(), new Integer(0));//new Integer(yPos));
      newProp.setYLocationOffset(yLocationOffset);

      add(newLabel);
      add(newProp);
      newProp.setIsOutput(true);

      newLabel.setFont(mPropertyFont);
      newLabel.setForeground(Color.black);
      newLabel.setHorizontalAlignment(JLabel.RIGHT);
    }
    return newProp;
  }

/*    public Entity getEntity() {
        return mEntity;
    }*/

    public boolean isEnclosureNode() {
        return false;
    }

    public boolean isFunctionNode() {
        return false;
    }

    public boolean isTransformationNode() {
      return false;
    }
/*
    public void setSceneGraph(SceneGraphPanel sceneGraph) {
        super.setSceneGraph(sceneGraph);
        for (int i = 0; i < mPropertyNodes.length; i++) {
            mPropertyNodes[i].setSceneGraph(sceneGraph);
        }
    }
*/
    private void jbInit() throws Exception {
        //this.setBackground(Color.green);
        mNameEdit = new NameEdit();
        add(mNameEdit);
        mNameEdit.setVisible(false);

        mDummyField = new JTextField();
        mDummyField.setBounds(0,0, 1, 1);
        mDummyField.setVisible(false);
        add(mDummyField);

        mInstanceOfLabel = new JLabelNoAntialias();
        mInstanceOfLabel.setFont(mPropertyFont);
        mInstanceOfLabel.setVisible(false);
        add(mInstanceOfLabel);

        mName = new JLabel();
        mName.addMouseListener(new MouseAdapter() {
          public void mouseClicked(MouseEvent e) {
            if (e.getClickCount() == 2) {
              editName();
            }
            else {
              e.translatePoint((int)mNameEdit.getLocation().getX() + 10,
                               (int)mNameEdit.getLocation().getY() + 10);
              mSelf.dispatchEvent(e);
            }
          }
          public void mousePressed(MouseEvent e) {
            e.translatePoint((int)mNameEdit.getLocation().getX() + 10,
                             (int)mNameEdit.getLocation().getY() + 10);
            mSelf.dispatchEvent(e);
          }
          public void mouseReleased(MouseEvent e) {
            e.translatePoint((int)mNameEdit.getLocation().getX() + 10,
                             (int)mNameEdit.getLocation().getY() + 10);
            mSelf.dispatchEvent(e);
          }
        });
        mName.addMouseMotionListener(new MouseMotionAdapter() {
          public void mouseDragged(MouseEvent e) {
            e.translatePoint((int)mNameEdit.getLocation().getX() + 10,
                             (int)mNameEdit.getLocation().getY() + 10);
            mSelf.dispatchEvent(e);
          }
          public void mouseMoved(MouseEvent e) {
            e.translatePoint((int)mNameEdit.getLocation().getX() + 10,
                             (int)mNameEdit.getLocation().getY() + 10);
            mSelf.dispatchEvent(e);
          }
        });


        add(mName);
        mName.setBounds(X_INSET + BORDER_THICKNESS + TITLE_X_OFFSET, Y_INSET + BORDER_THICKNESS,
                        getWidth() - (2 * (X_INSET + BORDER_THICKNESS) + TITLE_X_OFFSET), 16);
    }

    private void editName() {
      mName.setVisible(false);
      mNameEdit.setVisible(true);
      mNameEdit.setLocation(X_INSET + BORDER_THICKNESS + TITLE_X_OFFSET, Y_INSET + BORDER_THICKNESS);

      mNameEdit.setText(mName.getText());
      mNameEdit.requestFocus();
      mNameEdit.selectAll();
      validate();
    }

    public void endNameEdit() {
      mNameEdit.setVisible(false);
      setName(mEntity.mName);
      mName.setVisible(true);
      validate();
    }

    public class NameEdit extends JTextFieldNoAntialias {

      public NameEdit() {
        final NameEdit nameEdit = this;
        addFocusListener(new FocusAdapter() {
          public void focusGained(FocusEvent e) {
            selectAll();
            mLastFocus = nameEdit;
          }
        });

        addKeyListener(new KeyAdapter() {
          public void keyPressed(KeyEvent e) {
            if (e.getKeyCode() == KeyEvent.VK_ENTER || e.getKeyCode() == KeyEvent.VK_TAB) {
              if (update()) {
                endNameEdit();
              }
              else {
                setBackground(Color.red);
                selectAll();
              }
            }
          }
        });
      }

      public boolean update() {
        if (JNaming.isValidJavaIdentifier(getText())) {
          if (mEntity.mEnclosure != null)
            mEntity.mName = mEntity.mEnclosure.newName(getText());
          else
            mEntity.mName = getText();
          return true;
        }
        return false;
      }

    }

    public int getMinimumWidth() {
        return MINIMUM_WIDTH;
    }

    public int getMinimumHeight() {
        if (mEntity != null && mEntity.mProperties.length > 0) {
            int result = mEntity.mProperties.length * (PropertyNode.DIAMETER + MIN_PROPERTY_SEPARATION) +
                   getTitleBarHeight() + CONTROL_BAR_HEIGHT + 2 * BORDER_THICKNESS;
          return result;
        }
        return DEFAULT_HEIGHT;
    }

    protected void parseClick(MouseEvent e) {
        int x = e.getX();
        int y = e.getY();

        mDragMode = NONE;
        if (mOuterBounds.contains(x, y) && !mInnerBounds.contains(x, y)) {  // Clicked in border
            if (x <= X_INSET + CORNER_DRAG_OFFSET) {                // Left side
                if (y <= Y_INSET + CORNER_DRAG_OFFSET)
                    mDragMode = TOP_LEFT_CORNER;
                else if (y >= getHeight() - (Y_INSET + CORNER_DRAG_OFFSET))
                    mDragMode = BOTTOM_LEFT_CORNER;
                else
                    mDragMode = LEFT_SIDE;
            }
            else if (x >= getWidth() - (X_INSET + CORNER_DRAG_OFFSET)) {    // RIGHT SIDE
                if (y <= Y_INSET + CORNER_DRAG_OFFSET)
                    mDragMode = TOP_RIGHT_CORNER;
                else if (y >= getHeight() - (Y_INSET + CORNER_DRAG_OFFSET))
                    mDragMode = BOTTOM_RIGHT_CORNER;
                else
                    mDragMode = RIGHT_SIDE;
            }
            else if (y <= Y_INSET + CORNER_DRAG_OFFSET) {
                mDragMode = TOP_SIDE;
            }
            else if (y >= getHeight() - (Y_INSET + CORNER_DRAG_OFFSET)) {
                mDragMode = BOTTOM_SIDE;
            }
        }
    }

    public void clearConnections() {
        for (int i = 0; i < mPropertyNodes.length; i++) {
            mPropertyNodes[i].clearConnections();
        }
    }

  public void editPropNode(PropertyNode propNode) {
    if (mEntity.isFunction()) {
      if (((FunctionNode)this).mIsEditing)
        return;
    }
    if (mPropertyEditPanel != null) {   // One's already been made
      // selected same property as last time
      if (mPropertyEditPanel.mPropertyNode.mProperty == propNode.mProperty) {
        if (mEditingProperty) {
          remove(mPropertyEditPanel);
          showPropertyLabels(true);
          mEditingProperty = false;
//          if (propNode.mProperty.isFreeFloating()) {
//            propNode.setLocation((int)propNode.mFreeFloatingLocation.getX(), (int)propNode.mFreeFloatingLocation.getY());
//          }
          return;
        }
      }
      remove(mPropertyEditPanel);
    }

//    if (propNode.mProperty.isFreeFloating()) {
//      propNode.setLocation(23, 23);
//    }
    mPropertyEditPanel = new PropertyEditPanel(this, propNode);
    add(mPropertyEditPanel);
    mEditingProperty = true;
    mPropertyEditPanel.start();
    showPropertyLabels(false);
  }

  public void showPropertyLabels(boolean show) {
    for (int i = 0; i < mPropertyNodes.length; i++) {
      if (mPropertyNodes[i].getLabel() != null)
        mPropertyNodes[i].getLabel().setVisible(show);
    }
  }
    public class ResourceChooser extends JFileChooser {
      ResourceChooser() {
      }

      public boolean isTraversable(File f) {
        if (f == null)
          return false;
        if (f.getPath().startsWith(g.resourceDirectory.getPath())) {
          return super.isTraversable(f);
        }
        else return false;
      }


      public void setCurrentDirectory(File dir) {
        if (dir == null)
          return;
        if (dir.getPath().startsWith(g.resourceDirectory.getPath())) {
          super.setCurrentDirectory(dir);
        }
        else super.setCurrentDirectory(g.resourceDirectory);
      }
  }

  public void setIsInstance(boolean isInstance) {
    if (isInstance) {
      mInstanceOfLabel.setText("Instance of \"" + mEntity.mRRL.relativePath() + "\"");
      mInstanceOfLabel.setVisible(true);
    }
    else {
      mInstanceOfLabel.setVisible(false);
    }
  }

  protected File storeAs() {
    ResourceChooser chooser = new ResourceChooser();
   // Note: source for ExampleFileFilter can be found in FileChooserDemo,
   // under the demo/jfc directory in the Java 2 SDK, Standard Edition.
    ExampleFileFilter filter = new ExampleFileFilter();
    filter.addExtension("fun");
    filter.addExtension("enc");
    filter.setDescription("Roebling resources");
    chooser.setFileFilter(filter);
    chooser.setFileView(new ImageFileView());
    chooser.setFileSystemView(new ResourceFileSystemView(chooser.getFileSystemView()));
    chooser.setAcceptAllFileFilterUsed(false);
    chooser.setCurrentDirectory(g.resourceDirectory);
    int returnVal = chooser.showSaveDialog(g.sceneGraphFrame);
    if(returnVal == JFileChooser.APPROVE_OPTION) {
      return chooser.getSelectedFile();
    }
    else
      return null;

  }

  protected boolean serializeNode(XMLWriter xmlOut) {
    EnclosureNode encNode;
    Entity[]      directory;
    EntityNode    nextNode;

    try {
      xmlOut.startElement("entityNode");
      //Util.serializeElement(xmlOut, "entityID", mEntity.mID - g.fileWritingEntityIDOffset);
      Util.serializeElement(xmlOut, "left", getLocation().getX());
      Util.serializeElement(xmlOut, "top", getLocation().getY());
      Util.serializeElement(xmlOut, "width", getWidth());
      Util.serializeElement(xmlOut, "height", getHeight());
      if (isEnclosureNode()) {
        encNode = (EnclosureNode)this;
//        directory = ((Enclosure)encNode.mEntity).mDirectory;
        Util.serializeElement(xmlOut, "revealed", encNode.isRevealed());
//        for (int i = 0; i < directory.length; i++) {
//          nextNode = g.sceneGraphPanel.getEntityNode(directory[i]);
//          if (nextNode != null) {
//            nextNode.serializeNode(xmlOut);
//          }
//        }
      }

//      serializeProperties(xmlOut);      // Don't serialize prop nodes for now
      xmlOut.endElement("entityNode");
    }
    catch (Exception e) {
      return false;
    }
    return true;
  }
}