package roebling;

import java.awt.Color;

/**
 * Title:        Roebling
 * Description:  Parametric design tool.
 * Copyright:    Copyright (c) 2001
 * Company:      NewBricks
 * @author Simon Greenwold
 * @version 1.0
 */

import javax.swing.*;
import java.util.*;
import java.net.*;
import java.awt.event.*;
import java.awt.*;
import java.io.*;
import javax.swing.filechooser.*;

public class FunctionNode extends EntityNode {

    private static final int INTERIOR_X_OFFSET = 8;
    private static final int INTERIOR_TOP_OFFSET = 22;
    private static final int INTERIOR_BOTTOM_OFFSET = 17;
    private static final int FUNCTION_MINIMUM_HEIGHT = 100;

    private static final Color FUNCTION_COLOR = new Color(204, 61, 10, 60);
    private static final Color FUNCTION_INSTANCE_COLOR = new Color(254, 161, 10, 60);
    private static final Color DISABLED_TEXT_COLOR = new Color(230, 230, 230);
    private JToggleButton mEditCompileButton;
    private JButton mStoreButton;
    private JButton mRunButton;

    private JTextArea mPropertySynopsis1 = new JTextAreaNoAntialias();
    private JTextArea mPropertySynopsis2 = new JTextAreaNoAntialias();

    private JTabbedPane mTabPane = new JTabbedPaneNoAntialias(JTabbedPane.BOTTOM);

    private JPanel mCodeEditPanel = new JPanel(new BorderLayout());
    private JScrollPane mCodeEditScrollPane = new JScrollPane();
    private JTextArea mCodeEditTextArea = new UndoableTextArea();

    private JPanel mClassLevelCodeEditPanel = new JPanel(new BorderLayout());
    private JScrollPane mClassLevelCodeEditScrollPane = new JScrollPane();
    private JTextArea mClassLevelCodeEditTextArea = new UndoableTextArea();

    private JPanel mTriggerPanel = new JPanel(new GridLayout(3,1));
    private ButtonGroup mTriggerButtonGroup = new ButtonGroup();
    private JRadioButton mStandardTriggerButton = new JRadioButtonNoAntialias("Standard (input changed)");
    private JRadioButton mTimerTriggerButton = new JRadioButtonNoAntialias("Timer      Delay time:");
    private JTextField mTimerTriggerDelayField = new JTextFieldNoAntialias();
    private JRadioButton mSceneGraphButtonTriggerButton = new JRadioButtonNoAntialias("Scene graph button");

    //private JScrollPane mErrorTextScrollPane1 = new JScrollPane();
    private JTextArea mErrorText1 = new JTextAreaNoAntialias();
    //private JScrollPane mErrorTextScrollPane2 = new JScrollPane();
    private JTextArea mErrorText2 = new JTextAreaNoAntialias();
    public boolean mIsEditing;
    private Font mFunctionFont;


    public FunctionNode() {
        if (mFunctionFont == null)
          mFunctionFont = new Font("Monospaced", Font.PLAIN, 12);
        addButtons();
        initialize();

        addComponentListener(new ComponentAdapter() {
            public void componentResized(ComponentEvent e) {
                if (mIsEditing) {
                    mTabPane.setSize(getWidth() - 2 * INTERIOR_X_OFFSET, getHeight() - (INTERIOR_TOP_OFFSET + INTERIOR_BOTTOM_OFFSET));
                }
            }
        });

    }

    private void initialize() {
        mCodeEditPanel.setName("Function Code");
        mPropertySynopsis1.setEnabled(false);
        mPropertySynopsis1.setFont(mFunctionFont);
        mPropertySynopsis1.setBackground(DISABLED_TEXT_COLOR);
        mCodeEditPanel.add(mPropertySynopsis1, BorderLayout.NORTH);
        mCodeEditPanel.add(mCodeEditScrollPane, BorderLayout.CENTER);
        mCodeEditScrollPane.getViewport().add(mCodeEditTextArea);
        mCodeEditTextArea.setFont(mFunctionFont);
        CodeEditorListener codeEditListener = new CodeEditorListener(mCodeEditTextArea);
        mCodeEditTextArea.addKeyListener(codeEditListener);
        mCodeEditTextArea.addFocusListener(codeEditListener);
        //mCodeEditPanel.add(mErrorTextScrollPane1, BorderLayout.SOUTH);
        //mErrorTextScrollPane1.getViewport().add(mErrorText1);
        mCodeEditPanel.add(mErrorText1, BorderLayout.SOUTH);
        mErrorText1.setEditable(false);
        mErrorText1.setFont(mFunctionFont);

        mClassLevelCodeEditPanel.setName("Class Level Code");
        mPropertySynopsis2.setEnabled(false);
        mPropertySynopsis2.setFont(mFunctionFont);
        mPropertySynopsis2.setBackground(DISABLED_TEXT_COLOR);
        mClassLevelCodeEditPanel.add(mPropertySynopsis2, BorderLayout.NORTH);
        mClassLevelCodeEditPanel.add(mClassLevelCodeEditScrollPane, BorderLayout.CENTER);
        mClassLevelCodeEditScrollPane.getViewport().add(mClassLevelCodeEditTextArea);
        mClassLevelCodeEditTextArea.setFont(mFunctionFont);
        codeEditListener = new CodeEditorListener(mClassLevelCodeEditTextArea);
        mClassLevelCodeEditTextArea.addKeyListener(codeEditListener);
        mClassLevelCodeEditTextArea.addFocusListener(codeEditListener);
        //mErrorTextScrollPane2.getViewport().add(mErrorText2);
        //mClassLevelCodeEditPanel.add(mErrorTextScrollPane2, BorderLayout.SOUTH);
        mClassLevelCodeEditPanel.add(mErrorText2, BorderLayout.SOUTH);
        mErrorText2.setEditable(false);
        mErrorText2.setFont(mFunctionFont);

        mTriggerPanel.setName("Trigger type");
        mTriggerPanel.setBackground(Color.white);
        mTriggerButtonGroup.add(mStandardTriggerButton);
        mTriggerButtonGroup.add(mTimerTriggerButton);
        mTriggerButtonGroup.add(mSceneGraphButtonTriggerButton);

        JPanel tempPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
        tempPanel.setBackground(Color.white);
        mStandardTriggerButton.setBackground(Color.white);
        tempPanel.add(mStandardTriggerButton);
        mTriggerPanel.add(tempPanel);
        mTriggerPanel.add(tempPanel);

        tempPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
        tempPanel.setBackground(Color.white);
        mTimerTriggerButton.setBackground(Color.white);
        tempPanel.add(mTimerTriggerButton);
        tempPanel.add(mTimerTriggerDelayField);
        mTriggerPanel.add(tempPanel);

        tempPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
        tempPanel.setBackground(Color.white);
        mSceneGraphButtonTriggerButton.setBackground(Color.white);
        tempPanel.add(mSceneGraphButtonTriggerButton);
        mTriggerPanel.add(tempPanel);


        mTabPane.add(mCodeEditPanel);
        mTabPane.add(mClassLevelCodeEditPanel);
        mTabPane.add(mTriggerPanel);

        mRunButton = new JButtonNoAntialias();
        mRunButton.setText("Run");
        mRunButton.addActionListener(new ActionListener() {
          public void actionPerformed(ActionEvent e) {
            ((Function)mEntity).mRunOnce = true;
            mEntity.needsRefresh();
            mEntity.update();
            g.scene.refresh();
          }
        });
        //mRunButton.setVisible(false);
        mRunButton.setSize(60, 20);
        add(mRunButton);

//        mTabPane.add(mConstraintCodeEditPanel);
    }

    public void setEntity(Entity entity) {
      super.setEntity(entity);
      if (((Function)entity).mTriggerType == Function.SCENEGRAPH_TRIGGER) {
        mRunButton.setVisible(true);
      }
      else {
        mRunButton.setVisible(false);
      }
    }

    public boolean isFunctionNode() {
        return true;
    }

    public Color getFillColor() {
        if (! mEntity.mIsInstance)
          return FUNCTION_COLOR;
        return FUNCTION_INSTANCE_COLOR;
    }

    public int getMinimumHeight() {
        return (int)java.lang.Math.max(super.getMinimumHeight(), FUNCTION_MINIMUM_HEIGHT);
    }

    public Function getFunction() {
        return (Function)mEntity;
    }

    public void resized() {
        super.resized();

        mEditCompileButton.setBounds(20, getHeight() - 17, 36, 12);
        mStoreButton.setBounds(56, getHeight() - 17, 30, 12);
        mRunButton.setLocation((getWidth() - mRunButton.getWidth()) / 2, (getHeight() - mRunButton.getHeight()) / 2);

        if (mLastFocus != null) {
          mLastFocus.requestFocus();
        }

    }

    /*
    public void setEntity(Entity entity) {
//        ((Function)entity).createPropertyNodes();
        super.setEntity(entity);
    }
*/

    private boolean compile() {
        String newName = mEntity.mEnclosure.newName(mEntity.mName);   // Necessary?
        Class functionClass = Compiler.makeNewFunctionType(mEntity, mEntity.mName,
                                              mEntity.mProperties,
                                              mCodeEditTextArea.getText(),
                                              mClassLevelCodeEditTextArea.getText());  // Class level code

        String errorString = Compiler.getError();
        mErrorText1.setText(errorString);
        mErrorText2.setText(errorString);
        if (errorString.equals("")) {
          mErrorText1.setBackground(Color.white);
          mErrorText2.setBackground(Color.white);
        }
        else {
          System.out.println(Compiler.errorString);
          mErrorText1.setBackground(Color.pink);
          mErrorText2.setBackground(Color.pink);
        }

        if (functionClass == null)
            return false;

//        FunctionDirectoryEntry entry = new FunctionDirectoryEntry(newName,
//                                                    new File(g.tempDirectory + g.FS + newName + ".class"),
//                                                    functionClass);

        Function newFunc = g.scene.newFunction(functionClass, newName, mEntity.mEnclosure);
        EntityNode node = g.sceneGraphPanel.addNewEntityNode(newFunc, mEntity.mEnclosure);

        copyConnections(newFunc, mEntity);
        mEntity.delete();
        node.setBounds(getBounds());
        node.resized();

        mEnclosureNode.removeGraphNode(this);

        newFunc.mName = mEntity.mName;
        node.setName(newFunc.mName);

        g.sceneGraphPanel.setSelected(node);

        g.scene.refresh();
        return true;
    }

    private void copyConnections(Entity newEntity, Entity oldEntity) {
      Property newProp;
      Property oldProp;

      for (int i = 0; i < newEntity.mProperties.length; i++) {
        newProp = newEntity.mProperties[i];
        oldProp = oldEntity.mProperties[i];
        PropertyNode child, parent;


        if (oldProp.mParent != null) {
          boolean test = Property.newRelationship(oldProp.mParent, newProp);
          child = g.sceneGraphPanel.getInputNode(newProp);
          parent = g.sceneGraphPanel.getOutputNode(oldProp.mParent);

          child.newParent(parent);
          if (child.isEnclosureOut()) {
              parent.getEnclosureNode().newConnection(parent, child);
              parent.getEnclosureNode().repaint();
          }
          else {
              child.getEnclosureNode().newConnection(parent, child);
              child.getEnclosureNode().repaint();
          }

        }
        Property[] oldChildren = (Property[])oldProp.mChildren.clone();
        for (int childIndex = 0; childIndex < oldChildren.length; childIndex++) {
          Property.newRelationship(newProp, oldChildren[childIndex]);
          child = g.sceneGraphPanel.getInputNode(oldChildren[childIndex]);
          parent = g.sceneGraphPanel.getOutputNode(newProp);

          child.newParent(parent);
          if (child.isEnclosureOut()) {
              parent.getEnclosureNode().newConnection(parent, child);
              parent.getEnclosureNode().repaint();
          }
          else {
              child.getEnclosureNode().newConnection(parent, child);
              child.getEnclosureNode().repaint();
          }
        }

      }

    }

    private boolean exitCodeEdit() {

        if (mTimerTriggerButton.isSelected()) {
          ((Function)mEntity).mTriggerType = Function.TIMER_TRIGGER;
          try {
            ((Function)mEntity).mTimerInterval = Integer.parseInt(mTimerTriggerDelayField.getText().trim());
          }
          catch (Exception e) {
            ((Function)mEntity).mTimerInterval = Function.DEFAULT_TIMER_INTERVAL;
          }
        }
        else if (mSceneGraphButtonTriggerButton.isSelected()) {
          ((Function)mEntity).mTriggerType = Function.SCENEGRAPH_TRIGGER;
        }
        else {
          ((Function)mEntity).mTriggerType = Function.STANDARD_TRIGGER;
        }
        if (!compile())
            return false;

        mStoreButton.setVisible(true);
        mIsEditing = false;
        remove(mTabPane);
        showPropertyLabels(true);
        repaint();
        return true;
    }

    public void showPropertyLabels(boolean show) {
        for (int i = 0; i < mPropertyNodes.length; i++) {
            if (mPropertyNodes[i].getLabel() != null)
                mPropertyNodes[i].getLabel().setVisible(show);
        }
    }

    private void enterCodeEdit() {

      mRunButton.setVisible(false);
      mTimerTriggerDelayField.setText(" " + String.valueOf(((Function)mEntity).mTimerInterval));
      if (((Function)mEntity).mTriggerType == Function.SCENEGRAPH_TRIGGER) {
        mSceneGraphButtonTriggerButton.setSelected(true);
      }
      else if (((Function)mEntity).mTriggerType == Function.TIMER_TRIGGER) {
        mTimerTriggerButton.setSelected(true);
      }
      else {
        mStandardTriggerButton.setSelected(true);
      }

      mInstanceOfLabel.setVisible(false);
        if (mEditingProperty) {     // stop editing properties
          editPropNode(mPropertyEditPanel.mPropertyNode);
        }

        mIsEditing = true;
        mStoreButton.setVisible(false);
        showPropertyLabels(false);
        String propertySynopsis = makePropertySynopsis();
        mPropertySynopsis1.setText(propertySynopsis);
        mPropertySynopsis2.setText(propertySynopsis);

        mCodeEditTextArea.setText(getFunction().getCode());
        try {
            mClassLevelCodeEditTextArea.setText(getFunction().getClassLevelCode());
        }
        catch (Exception e) {
            mClassLevelCodeEditTextArea.setText("");
        }
/*        try {
            mConstraintCodeEditTextArea.setText(getFunction().getConstraintCode());
        }
        catch (Exception e) {
            mConstraintCodeEditTextArea.setText("");
        }*/

        add(mTabPane);
        mTabPane.setBounds(INTERIOR_X_OFFSET, INTERIOR_TOP_OFFSET,
                            getWidth() - 2 * INTERIOR_X_OFFSET, getHeight() - (INTERIOR_TOP_OFFSET + INTERIOR_BOTTOM_OFFSET));
        repaint();
    }
/*
    private void showPropertyLabels(boolean show) {
        for (int i = 0; i < mPropertyNodes.length; i++) {
            if (mPropertyNodes[i].getLabel() != null)
                mPropertyNodes[i].getLabel().setVisible(show);
        }
    }
*/
    private String makePropertySynopsis() {
        Property[] properties = mEntity.mProperties;
        String propString;
        String arrayString;

        propString = "// INPUT Properties: \n";
        for (int i = 0; i < properties.length; i++) {
            if (! properties[i].isInput())
              continue;
            if (properties[i].isArray)
              arrayString = "[]";
            else
              arrayString = "";
            if (properties[i].mType == Property.NUMBER) {
                propString += "  double" + arrayString + " ";
            }
            if (properties[i].mType == Property.LOCATION) {
                propString += "  LocationProp ";
            }
            if (properties[i].mType == Property.STRING) {
                propString += "  String" + arrayString + " ";
            }
            if (properties[i].mType == Property.BOOLEAN) {
                propString += "  boolean" + arrayString + " ";
            }
            if (properties[i].mType == Property.COLOR) {
                propString += "  ColorProp ";
            }
            propString += properties[i].mName + ";   ";
            if (properties[i].isCompound && properties[i].isArray) {
              propString += "// ARRAY\n";
            }
            else {
              propString += "\n";
            }
        }
        propString += "// OUTPUT Properties: \n";
        for (int i = 0; i < properties.length; i++) {
            if (! properties[i].isOutput())
              continue;
            if (properties[i].isArray)
              arrayString = "[]";
            else
              arrayString = "";

            if (properties[i].mType == Property.NUMBER) {
                propString += "  double" + arrayString + " ";
            }
            if (properties[i].mType == Property.LOCATION) {
                propString += "  LocationProp ";
            }
            if (properties[i].mType == Property.STRING) {
                propString += "  String" + arrayString + " ";
            }
            if (properties[i].mType == Property.BOOLEAN) {
                propString += "  boolean" + arrayString + " ";
            }
            if (properties[i].mType == Property.COLOR) {
                propString += "  ColorProp ";
            }
            propString += properties[i].mName + ";   ";
            if (properties[i].isCompound && properties[i].isArray) {
              propString += "// ARRAY\n";
            }
            else {
              propString += "\n";
            }
        }

        return propString;
    }

    private void addButtons() {
        String jarPath = "";
        String tempStr;
        try {
            StringTokenizer pathTokens = new StringTokenizer(System.getProperty("java.class.path"), ";");
            while (pathTokens.hasMoreTokens()) {
                tempStr = pathTokens.nextToken();
                if (tempStr.indexOf("paramour.jar") != -1) jarPath = tempStr;
            }
        }
        catch (Exception e) {
        }

        if (!jarPath.equals("")) {
            jarPath = "jar:file:" + jarPath + "!/";
        }
        else
            jarPath = "file:";

        try {
            mEditCompileButton = new JToggleButton(new ImageIcon(new URL(jarPath + "images/editCode.gif")));
            mEditCompileButton.setSelectedIcon(new ImageIcon(new URL(jarPath + "images/Compile.gif")));
            mEditCompileButton.addActionListener(new ActionListener()  {
                public void actionPerformed(ActionEvent e) {
                    if (mEditCompileButton.isSelected()) {
                        enterCodeEdit();
                    }
                    else {
                        if (!exitCodeEdit()) {
                            mEditCompileButton.setSelected(true);
                        }
                    }
                }
            });
            mEditCompileButton.setSelected(false);
            add(mEditCompileButton);

            mStoreButton = new JButton(new ImageIcon(new URL(jarPath + "images/Store.gif")));
            mStoreButton.addActionListener(new ActionListener()  {
                public void actionPerformed(ActionEvent e) {
                    storeFunction();
                }
            });
            add(mStoreButton);
        }
        catch (Exception e) {
        }
    }

    public class JRadioButtonNoAntialias extends JRadioButton {
        public JRadioButtonNoAntialias(String text) {
            super(text);
        }
        public void paintComponent(Graphics g) {
            Graphics2D g2 = (Graphics2D)g;
            g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF);
            super.paintComponent(g2);
        }
    }

    public class JTabbedPaneNoAntialias extends JTabbedPane {
        public JTabbedPaneNoAntialias(int tabLocation) {
            super(tabLocation);
        }
        public void paintComponent(Graphics g) {
            Graphics2D g2 = (Graphics2D)g;
            g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF);
            super.paintComponent(g2);
        }
    }

/*    public class ExampleFileFilter extends javax.swing.filechooser.FileFilter {
      ExampleFileFilter() {
      }
      public boolean accept(File f) {
        if (f.isDirectory()) return true;
        String type;
        type = FileUtil.getPostfix(f.getName());

        return ( (type == "enc" || type == "fun") );
      }
      public String getDescription() {
        return "Roebling resources.";
      }
    }*/


    private void storeFunction() {
        Function function = getFunction();
        File storeFile = null;
        Function newFunction = null;

        File chosenFile = storeAs();
        if (chosenFile == null)
          return;

        Class newClass = Compiler.makeNewFunctionType(function, function.mName,
                                                      function.mProperties,
                                                      function.getCode(),
                                                      function.getClassLevelCode());
        if (newClass == null)
            return;

        try {
            newFunction = ((Function)newClass.newInstance());
            storeFile = new File(g.tempDirectory + g.FS + "roebling" + g.FS +
                                 FileUtil.getPostfix(newClass.getName()) + ".class");
        }
        catch (Exception e) {
          return;
        }

        /*FunctionDirectoryEntry entry = g.functionsFrame.getSelectedStoredFunction();
        if (entry == null) {
            entry = g.functionsFrame.getFunctionRoot();
        }
        else if (! entry.isDirectory()) {
            entry = (FunctionDirectoryEntry)entry.getParent();
        }*/

        String fileName = chosenFile.getParent() + g.FS + FileUtil.stripPostfix(chosenFile.getName()) + ".fun";
        File newFile = new File(fileName);

        try {
          FileUtil.zipUp(new String[] {storeFile.getPath()}, g.tempDirectory.getPath(),fileName);
        }
        catch (Exception e) {
        }

        g.resourcesFrame.reloadResourceTree();

        mEntity.mIsInstance = true;
        mEntity.mRRL = g.resourcesFrame.getRRL(newFile);
        setIsInstance(true);
        repaint();
      }

// Adapted from Processing code by Ben Fry
  public class CodeEditorListener extends KeyAdapter implements FocusListener
  {
    static final String spaces = "                                    ";
    String tabString;
    String newline = System.getProperty("line.separator");


    boolean expandTabs;
    int tabSize;
    boolean autoIndent;


    boolean balanceParens;
    boolean balancing = false;
    JTextArea tc;
    int selectionStart, selectionEnd;
    int position;



    public CodeEditorListener(JTextArea textComponent) {
      expandTabs = true;                         //PdeApplet.getBoolean("editor.expandTabs", false);
      tabSize = 2;                            //PdeApplet.getInteger("editor.tabSize", 2);
      tabString = spaces.substring(0, tabSize);
      autoIndent = true;                      //PdeApplet.getBoolean("editor.autoIndent", false);
      balanceParens = false;            //PdeApplet.getBoolean("editor.balanceParens", false);
      tc = textComponent;
    }



    public void keyPressed(KeyEvent event) {
      // only works with TextArea, because it needs 'insert'
      //TextComponent tc = (TextComponent) event.getSource();
//      tc = (JTextArea) event.getSource();
      deselect();
      char c = event.getKeyChar();


      //System.err.println((int)c);
      switch ((int) c) {
      case ')':
        if (balanceParens) {
          position = tc.getCaretPosition() + 1;
          char contents[] = tc.getText().toCharArray();
          int counter = 1; // char not in the textfield yet
          //int index = contents.length-1;
          int index = tc.getCaretPosition() - 1;
          boolean error = false;
          if (index == -1) {  // special case for first char
            counter = 0;
            error = true;
          }
          while (counter != 0) {
            if (contents[index] == ')') counter++;
            if (contents[index] == '(') counter--;
            index--;
            if ((index == -1) && (counter != 0)) {
              error = true;
              break;
            }
          }
          if (error) {
            //System.err.println("mismatched paren");
            Toolkit.getDefaultToolkit().beep();
            tc.select(0, 0);
            tc.setCaretPosition(position);
          }
          tc.insert(")", position-1);
          event.consume();
          if (!error) {
            selectionStart = index+1;
            selectionEnd = index+2;
            tc.select(selectionStart, selectionEnd);
            balancing = true;
          }
        }
        break;


      case 9:  // expand tabs
        if (expandTabs) {
          //System.out.println("start = " + tc.getSelectionStart());
          //System.out.println("end = " + tc.getSelectionEnd());
          //System.out.println("pos = " + tc.getCaretPosition());
          tc.replaceRange(tabString, tc.getSelectionStart(),
                          tc.getSelectionEnd());
          event.consume();
        }
        break;


      case 10:  // auto-indent
        if (autoIndent) {
          //System.err.println("auto indenting");
          char contents[] = tc.getText().toCharArray();
          // back up until \r \r\n or \n.. @#($* cross platform
          //index = contents.length-1;
          int index = tc.getCaretPosition() - 1;
          int spaceCount = 0;
          boolean finished = false;
          while ((index != -1) && (!finished)) {
            if ((contents[index] == '\r') ||
                (contents[index] == '\n')) {
              finished = true;
            } else {
              spaceCount = (contents[index] == ' ') ?
                (spaceCount + 1) : 0;
            }
            index--;
          }


          // !@#$@#$ MS VM doesn't move the caret position to the
          // end of an insertion after it happens, even though sun does
          String insertion = "\n" + spaces.substring(0, spaceCount); //newline + spaces.substring(0, spaceCount);
          int oldCarrot = tc.getSelectionStart();
          tc.replaceRange(insertion, oldCarrot, tc.getSelectionEnd());
          // microsoft vm version:
          //tc.setCaretPosition(oldCarrot + insertion.length() - 1);
          // sun vm version:
          tc.setCaretPosition(oldCarrot + insertion.length());
          event.consume();
        }
        break;


      case 1: tc.selectAll(); break;  // control a for select all
      }
    }



    protected void deselect() {
      if (!balancing || (tc == null)) return;
      // bounce back, otherwise will write over stuff
      if ((selectionStart == tc.getSelectionStart()) &&
          (selectionEnd == tc.getSelectionEnd()))
        tc.setCaretPosition(position);
      balancing = false;
    }



    public void focusGained(FocusEvent event) {
      mLastFocus = tc;
    }



    public void focusLost(FocusEvent event) {
      deselect();
    }
  }



}