package roebling;

/**
 * Title:        Roebling
 * Description:  Parametric dataflow programming.
 * Copyright:    Copyright (c) 2001
 * Company:
 * @author Simon Greenwold
 * @version 1.0
 */

// Questestion: How do we freeze and restore instancing information between runs?
// We'd need to have some sort of referencing protocol that will not fail

// For functions, it's easy-- it's the class name. Those shouldn't be reused. They'll all be
// loaded at runtime, so they'll be available by name. It's perfect.

// For enclosures, it's harder-- it should be the path of the saved image of the enclosure.

// What if somebody instances an enclosure or function that has not been saved?
// Then it needs to be saved to a special directory of instance masters first. When it is saved
// later elsewhere, it will be moved. (Is that too annoying? Why not just require them to be
// saved first?

// Instance masters should not be deleteable unless they have no references.

// Let's say you save two versions of a scene with different names both of which contain
// masters of the same instances.
// Who dominates? Are they both masters? No. One can change. Is the one created second an
// instance of the first? Then if you try to edit it, you must edit the master instead or become
// the master.

// When instances are stored, all that needs to be stored is a reliable reference to the
// the master, and input property values and outside prop connections.

// This is a disaster. It's way too complicated. You must simplify:
// 1) No anonymous masters. Masters live among the regular folks. But they are all saved
// at the root level.

// How do you create an instance of something?
// 1) Pull it off a list (function or enclosure)
// 2) Click an instantiate button on a live function or enclosure
//    This will require you to store the function or enclosure before instantiating.

// How do you edit a master?
// 1) Create an instance. (Or refer to an existing instance.)
// 2) Attempt to edit it.
//    You will be asked if you want to 1) Break instance relationship & edit
//                                  or 2) Make this master and edit

// What if you make changes to a master and do not store them?
// (You only store them when an enclosure is stored.) Then the changes are absorbed
// when you save the enclosure. If you never store it, the changes are discarded.

// What if you change the name of a master and store it?
// It is a new master, and all references of the old master are updated to reflect that.

// How do we make changes to references to a master that are in closed files?
// Do we have to? Can such things be resolved on open? You'd have to save a record...

// NO! HERE'S THE LAW
// You cannot change the name of a master. A master saves over itself only. That way
// no matter where references are buried, they can simply...

// NO! HERE'S A BETTER WAY ENTIRELY
// We introduce a new level of annoying complexity. There is such a thing basically as
// an instance manager. It maintains, for a given NAME (is that name with path?), where
// the current master lives. That way whenever a reference is loaded it checks with this
// list and gets the most recent stuff.
// The only references that need to be changed on the fly are the ones that are currently
// open. These must be tracked.

// Then you can do some cool stuff. You can reassign mastership of a given slot to something else.
// You may never remove master slots. If you do, you risk stranding unknown references. (You
// could do a file sweep to detect these.)

// This means there's a given name for the master slot which you'd better not change.

// Look at it this way. All references find a master by this name on a board. What that
// actually attaches to is up to you.
// You can change it.

// So it's like this:
// REFERENCE  --> LOCATION
//  Triangle  -->  moof/dwingle/porf

// How does a saved thing know it is a master? It's a master if it isn't a reference.
// How does it know not to change names? Does it need to? Perhaps when any master changes
// names, it notifies the board and the board changes the reference (and reloads any
// open references).

//Can you unhitch a master from its role? No. You could make it a new master and then
// change the reference in the original to point to it.

// FINAL IMPLEMENTATION
// All instances are held in a single directory called "resources"
// There is no tracking other than by file name
// There's virutally no interface here except to ge able to get the RRL
// (Roebling Resource Locator) of the stored


public interface Instanceable {
  public boolean isInstance();
  public void setIsInstance(boolean isInstance);
  public void setResource(RRL resource);
  public RRL getResource();
  public void reloadFromResource();
}