package roebling;

/**
 * Title:        Roebling
 * Description:  Parametric dataflow programming.
 * Copyright:    Copyright (c) 2001
 * Company:
 * @author Simon Greenwold
 * @version 1.0
 */
import java.awt.*;
import java.awt.geom.*;

public class Line extends GraphicEntity {

  public final static int DEFAULT_SIZE = 2;

  private Line2D mPaintLine = new Line2D.Double();

  private NumProp mSizeProperty;
  private float mSize;
  private LocationProp mStartPointProperty;
//  private Point2D mStartPoint;
  private LocationProp mEndPointProperty;
//  private Point2D mEndPoint;
  private BoolProp mVisibleProperty;
  private BoolProp mSelectableProperty;
  private ColorProp mColorProperty;
  private Color mColor;
  private NumProp mLayerProperty;

  private BasicStroke mLineStroke;
  private Point2D mPreDragStartPoint = new Point2D.Double();
  private Point2D mPreDragEndPoint = new Point2D.Double();
  private Point2D mPostDragStartPoint = new Point2D.Double();
  private Point2D mPostDragEndPoint = new Point2D.Double();
  private Point2D.Double mPanelStartPoint = new Point2D.Double();
  private Point2D.Double mPanelEndPoint = new Point2D.Double();

  public Line() {
    super();
    mEntityType = "line";
    addProperties();
  }

  public Line(String name, Enclosure enclosure) {
    super(name, enclosure);
    mEntityType = "line";
    addProperties();
  }

  protected void addProperties() {
    mSizeProperty  = new NumProp();
    mSizeProperty.mName = "Size";
    mSizeProperty.value = DEFAULT_SIZE;
    mSizeProperty.valueSet();
    addProperty(mSizeProperty);

    mStartPointProperty = new LocationProp();
    mStartPointProperty.mName = "StartPoint";
    addProperty(mStartPointProperty);

    mEndPointProperty = new LocationProp();
    mEndPointProperty.mName = "EndPoint";
    addProperty(mEndPointProperty);

    mColorProperty = new ColorProp();
    mColorProperty.mName = "Color";
    mColorProperty.a = 255;
    addProperty(mColorProperty);

    mLayerProperty = new NumProp();
    mLayerProperty.mName = "Layer";
    mLayerProperty.value = 0;
    addProperty(mLayerProperty);

    mVisibleProperty = new BoolProp();
    mVisibleProperty.mName = "Visible";
    mVisibleProperty.value = true;
    addProperty(mVisibleProperty);

    mSelectableProperty = new BoolProp();
    mSelectableProperty.mName = "Selectable";
    mSelectableProperty.value = true;
    addProperty(mSelectableProperty);
  }

  public void setColor(double r, double g, double b) {
    setColor(r,g,b, 255);
  }
  public void setColor(double r, double g, double b, double a) {
    mColorProperty.r = r;
    mColorProperty.g = g;
    mColorProperty.b = b;
    mColorProperty.a = a;
    mColorProperty.valueSet();
  }

  public void setSize(double size) {
    mSizeProperty.value = size;
    mSizeProperty.valueSet();
  }

  public void setStartPoint(double x, double y) {
    mStartPointProperty.x = x;
    mStartPointProperty.y = y;
    mStartPointProperty.valueSet();
  }

  public void setStartPoint(Point2D startPoint) {
    mStartPointProperty.x = startPoint.getX();
    mStartPointProperty.y = startPoint.getY();
    mStartPointProperty.valueSet();
  }

  public void setEndPoint(double x, double y) {
    mEndPointProperty.x = x;
    mEndPointProperty.y = y;
    mEndPointProperty.valueSet();
  }

  public void setEndPoint(Point2D endPoint) {
    mEndPointProperty.x = endPoint.getX();
    mEndPointProperty.y = endPoint.getY();
    mEndPointProperty.valueSet();
  }

  protected void readInputs() {
    mSizeProperty.validate();
    mSize = (float)java.lang.Math.max(mSizeProperty.value, 0.0);
    mColorProperty.validate();
    mColor = mColorProperty.getColorValue();
    mStartPointProperty.validate();
    mPanelStartPoint.setLocation(g.transform.toPanelLocation(new Point2D.Double(mStartPointProperty.x, mStartPointProperty.y)));
    mEndPointProperty.validate();
    mPanelEndPoint.setLocation(g.transform.toPanelLocation(new Point2D.Double(mEndPointProperty.x, mEndPointProperty.y)));
    mVisibleProperty.validate();
    mVisible = mVisibleProperty.value;
    mSelectableProperty.validate();
    mSelectable = mSelectableProperty.value;
    mLayerProperty.validate();
    mLayer = (int)mLayerProperty.value;
  }

  public void draw(Graphics2D graphics) {
//    update();
    if (! mVisible)
        return;

    mEnclosure.makeTransform();
    mSaveTrans = graphics.getTransform();
    graphics.transform(mEnclosure.mCompositeTrans);

    graphics.setStroke(mLineStroke);
    graphics.setPaint(mColor);
    graphics.draw(mPaintLine);

    graphics.setTransform(mSaveTrans);
  }

  public void savePreDragLocation() {
    mPreDragStartPoint.setLocation(mPanelStartPoint);
    mPreDragEndPoint.setLocation(mPanelEndPoint);
  }

  public  boolean hit(Point2D hitLocation)  {
    if (!mSelectable)
      return false;
    mEnclosure.inverseTransform(hitLocation,mTransHitPoint);
    return (mPaintLine.ptLineDist(mTransHitPoint) <= (mSize / 2.0) + 2);

    //return mPaintLine.intersects(hitLocation.getX() - 2, hitLocation.getY() - 2, 4, 4);
  }

  public void move(Point2D mouseDelta) {
    mPostDragStartPoint.setLocation(mPreDragStartPoint.getX() + mouseDelta.getX(),
                                   mPreDragStartPoint.getY() + mouseDelta.getY());
    mPostDragEndPoint.setLocation(mPreDragEndPoint.getX() + mouseDelta.getX(),
                                   mPreDragEndPoint.getY() + mouseDelta.getY());

    mStartPointProperty.setLocation(g.transform.toGeomLocation(mPostDragStartPoint));
    mEndPointProperty.setLocation(g.transform.toGeomLocation(mPostDragEndPoint));
    mStartPointProperty.valueSet();
    mEndPointProperty.valueSet();
  }

  public void run() {
    mLineStroke = new BasicStroke(mSize, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND);
    mPaintLine.setLine(mPanelStartPoint, mPanelEndPoint);
  }
}