package roebling;

/**
 * Title:        Roebling
 * Description:  Parametric dataflow programming.
 * Copyright:    Copyright (c) 2001
 * Company:
 * @author Simon Greenwold
 * @version 1.0
 */
import java.awt.*;
import java.awt.geom.*;

public class Point extends GraphicEntity {

  public final static int DEFAULT_SIZE = 6;

  private Ellipse2D mPaintPoint = new Ellipse2D.Double();

  private NumProp mSizeProperty;
  private double mSize;
  private LocationProp mLocationProperty;
  private Point2D mLocation;
  private BoolProp mVisibleProperty;
  private BoolProp mSelectableProperty;
  private ColorProp mColorProperty;
  private Color mColor;
  private Point2D.Double mPanelLocation = new Point2D.Double();
  private NumProp mLayerProperty;

  private Point2D mPreDragLocation = new Point2D.Double();
  private Point2D mPostDragLocation = new Point2D.Double();

  public Point() {
    super();
    mEntityType = "point";
    addProperties();
  }

  public Point(String name, Enclosure enclosure) {
    super(name, enclosure);
    mEntityType = "point";
    addProperties();
  }

  protected void addProperties() {
      mSizeProperty  = new NumProp();
      mSizeProperty.mName = "Size";
      mSizeProperty.value = DEFAULT_SIZE;
      mSizeProperty.valueSet();
      addProperty(mSizeProperty);

      mLocationProperty = new LocationProp();
      mLocationProperty.mName = "Location";
      addProperty(mLocationProperty);

      mColorProperty = new ColorProp();
      mColorProperty.mName = "Color";
      //mColorProperty.r = 255;
      mColorProperty.a = 255;
      addProperty(mColorProperty);

    mLayerProperty = new NumProp();
    mLayerProperty.mName = "Layer";
    mLayerProperty.value = 0;
    addProperty(mLayerProperty);

      mVisibleProperty = new BoolProp();
      mVisibleProperty.mName = "Visible";
      mVisibleProperty.value = true;
      addProperty(mVisibleProperty);

    mSelectableProperty = new BoolProp();
    mSelectableProperty.mName = "Selectable";
    mSelectableProperty.value = true;
    addProperty(mSelectableProperty);
  }

  protected void readInputs() {
    mSizeProperty.validate();
    mSize = java.lang.Math.max(mSizeProperty.value, 0.0);
    mColorProperty.validate();
    mColor = mColorProperty.getColorValue();
    mLocationProperty.validate();
    mPanelLocation.setLocation(g.transform.toPanelLocation(new Point2D.Double(mLocationProperty.x, mLocationProperty.y)));
    mVisibleProperty.validate();
    mVisible = mVisibleProperty.value;
    mSelectableProperty.validate();
    mSelectable = mSelectableProperty.value;
    mLayerProperty.validate();
    mLayer = (int)mLayerProperty.value;
  }

  public void run() {
    mPaintPoint.setFrame((int)mPanelLocation.getX() - (mSize / 2),(int)mPanelLocation.getY() - (mSize / 2), mSize, mSize);
  }

  public void draw(Graphics2D graphics) {
//        System.out.println(mName + " drawing.");
//    update();

//        System.out.println("In Point draw: " + getName() + " VisibleProperty: " + mIsVisible.getBooleanValue() + " mVisible: " + mVisible);

    if (! mVisible)
      return;

    mEnclosure.makeTransform();
    mSaveTrans = graphics.getTransform();
    graphics.transform(mEnclosure.mCompositeTrans);

    graphics.setPaint(mColor);
    graphics.fill(mPaintPoint);

    graphics.setTransform(mSaveTrans);
  }

  public void setColor(double r, double g, double b) {
    setColor(r,g,b, 255);
  }
  public void setColor(double r, double g, double b, double a) {
    mColorProperty.r = r;
    mColorProperty.g = g;
    mColorProperty.b = b;
    mColorProperty.a = a;
    mColorProperty.valueSet();
  }

  public void setSize(double size) {
    mSizeProperty.value = size;
    mSizeProperty.valueSet();
  }

  public void setLocation(double x, double y) {
    mLocationProperty.x = x;
    mLocationProperty.y = y;
    mLocationProperty.valueSet();
  }

  public void setLocation(Point2D point) {
    mLocationProperty.x = point.getX();
    mLocationProperty.y = point.getY();
    mLocationProperty.valueSet();
  }

  public boolean hit(Point2D hitLocation) {
    if (!mSelectable)
      return false;
    mEnclosure.inverseTransform(hitLocation,mTransHitPoint);
    return mPaintPoint.intersects(mTransHitPoint.getX() - 2, mTransHitPoint.getY() - 2, 4, 4);
  }

  public void moveTo(Point2D newLocation) {
    mLocationProperty.x = g.transform.toGeomLocation(newLocation).getX();
    mLocationProperty.y = g.transform.toGeomLocation(newLocation).getY();
    mLocationProperty.valueSet();
  }

  public void move(Point2D mouseDelta) {
    mPostDragLocation.setLocation(mPreDragLocation.getX() + mouseDelta.getX(),
                                  mPreDragLocation.getY() + mouseDelta.getY());
    moveTo(mPostDragLocation);
  }

  public void savePreDragLocation() {
    mPreDragLocation.setLocation(mPanelLocation);
  }

}