package roebling;

/**
 * Title:        Roebling
 * Description:  Parametric dataflow programming.
 * Copyright:    Copyright (c) 2001
 * Company:
 * @author Simon Greenwold
 * @version 1.0
 */
import java.awt.*;
import java.awt.geom.*;

public class Polygon extends GraphicEntity {

  public final static int DEFAULT_THICKNESS = 2;

  private java.awt.Polygon mPolygon = new java.awt.Polygon();

  private NumProp mLineThicknessProperty;
  private float mLineThickness;
  private LocationProp mPointsProperty;

  private BoolProp mVisibleProperty;
  private BoolProp mSelectableProperty;
  private ColorProp mFillColorProperty;
  private Color mFillColor;
  private ColorProp mLineColorProperty;
  private Color mLineColor;
  private NumProp mLayerProperty;

  private BasicStroke mLineStroke;


  private Point2D.Double[] mPreDragPoints;
  private int[]  mPanelXs;
  private int[]  mPanelYs;
  /*
  private Point2D mPreDragStartPoint = new Point2D.Double();
  private Point2D mPreDragEndPoint = new Point2D.Double();
  private Point2D mPostDragStartPoint = new Point2D.Double();
  private Point2D mPostDragEndPoint = new Point2D.Double();
  private Point2D.Double mPanelStartPoint = new Point2D.Double();
  private Point2D.Double mPanelEndPoint = new Point2D.Double();
  */

  public Polygon() {
    super();
    mEntityType = "polygon";
    addProperties();
  }

  public Polygon(String name, Enclosure enclosure) {
    super(name, enclosure);
    mEntityType = "polygon";
    addProperties();
  }

  protected void addProperties() {
    mLineThicknessProperty  = new NumProp();
    mLineThicknessProperty.mName = "Stroke";
    mLineThicknessProperty.value = DEFAULT_THICKNESS;
    mLineThicknessProperty.valueSet();
    addProperty(mLineThicknessProperty);

    mPointsProperty = new LocationProp();
    mPointsProperty.mName = "Points";
    mPointsProperty.isArray = true;
    addProperty(mPointsProperty);

    mFillColorProperty = new ColorProp();
    mFillColorProperty.mName = "FillColor";
    mFillColorProperty.r = 255;
    mFillColorProperty.a = 80;
    addProperty(mFillColorProperty);

    mLineColorProperty = new ColorProp();
    mLineColorProperty.mName = "LineColor";
    mLineColorProperty.b = 255;
    mLineColorProperty.a = 255;
    addProperty(mLineColorProperty);

    mLayerProperty = new NumProp();
    mLayerProperty.mName = "Layer";
    mLayerProperty.value = 0;
    addProperty(mLayerProperty);

    mVisibleProperty = new BoolProp();
    mVisibleProperty.mName = "Visible";
    mVisibleProperty.value = true;
    addProperty(mVisibleProperty);

    mSelectableProperty = new BoolProp();
    mSelectableProperty.mName = "Selectable";
    mSelectableProperty.value = true;
    addProperty(mSelectableProperty);
  }

  public void addPoint(Point2D newPoint) {
    mPointsProperty.xs = Util.append(mPointsProperty.xs, newPoint.getX());
    mPointsProperty.ys = Util.append(mPointsProperty.ys, newPoint.getY());
    mPointsProperty.valueSet();
  }

  public void draw(Graphics2D graphics) {
//    update();
    if (! mVisible)
        return;

    mEnclosure.makeTransform();
    mSaveTrans = graphics.getTransform();
    graphics.transform(mEnclosure.mCompositeTrans);

    if (mPanelXs.length > 2) {
      graphics.setPaint(mFillColor);
      graphics.fill(mPolygon);
      graphics.setStroke(mLineStroke);
      graphics.setPaint(mLineColor);
      graphics.draw(mPolygon);
    }
    graphics.setTransform(mSaveTrans);
  }

  public  boolean hit(Point2D hitLocation)  {
    if (!mSelectable)
      return false;
    mEnclosure.inverseTransform(hitLocation,mTransHitPoint);
    return mPolygon.contains(mTransHitPoint);
  }

  public void savePreDragLocation() {
    mPreDragPoints = new Point2D.Double[mPointsProperty.xs.length];
    for (int i = 0; i < mPointsProperty.xs.length; i++) {
      mPreDragPoints[i] = new Point2D.Double(mPanelXs[i], mPanelYs[i]);
    }
  }

  public void move(Point2D mouseDelta) {
    for (int i = 0; i < mPointsProperty.xs.length; i++) {
      mPointsProperty.xs[i] = g.transform.xToGeom(mPreDragPoints[i].getX() + mouseDelta.getX());
      mPointsProperty.ys[i] = g.transform.yToGeom(mPreDragPoints[i].getY() + mouseDelta.getY());
    }
    mPointsProperty.valueSet();
  }

  public void run() {
    mLineStroke = new BasicStroke(mLineThickness, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND);
    if (mPanelXs.length > 2)
      mPolygon = new java.awt.Polygon(mPanelXs, mPanelYs, mPanelXs.length);
  }

  protected void readInputs() {
    mLineThicknessProperty.validate();
    mLineThickness = (float)java.lang.Math.max(mLineThicknessProperty.value, 0.0);
    mFillColorProperty.validate();
    mFillColor = mFillColorProperty.getColorValue();
    mLineColorProperty.validate();
    mLineColor = mLineColorProperty.getColorValue();
    mPointsProperty.validate();
    mPanelXs = new int[mPointsProperty.xs.length];
    mPanelYs = new int[mPointsProperty.ys.length];
    for (int i = 0; i < mPointsProperty.xs.length; i++) {
      mPanelXs[i] = (int)g.transform.xToPanel(mPointsProperty.xs[i]);
      mPanelYs[i] = (int)g.transform.yToPanel(mPointsProperty.ys[i]);
    }
    mVisibleProperty.validate();
    mVisible = mVisibleProperty.value;
    mSelectableProperty.validate();
    mSelectable = mSelectableProperty.value;
    mLayerProperty.validate();
    mLayer = (int)mLayerProperty.value;
  }
}