package roebling;

/**
 * Title:        Roebling
 * Description:  Parametric dataflow programming.
 * Copyright:    Copyright (c) 2001
 * Company:
 * @author Simon Greenwold
 * @version 1.0
 */

import java.util.*;
import java.awt.geom.*;
import java.awt.*;

public class Property {

  static int NUMBER;
  static int BOOLEAN ;
  static int STRING;
  static int LOCATION;
  static int COLOR;
  static int CUSTOM;

  public Entity mOwner;
  public String mName;
  public int mType;
  public int mID;

  public Property mParent = null;
  public Property mChildren[] = new Property[0];

  public boolean isArray;
  public boolean isCompound;

  public boolean mIsOutput;
  public boolean mIsInput;
  public boolean mIsEnclosureOut;
  public boolean mIsEnclosureIn;
  public boolean mIsFreeFloating;
  public Point2D mFreeFloatingLocationHint;

  public static HashMap mIDtoProperty = new HashMap();

  // update variables
  private int mGetCycle;
  private int mSetCycle;
  private boolean tempBoolean;
  private boolean mInIsSettable;

  public Property() {
    isArray = false;
    isCompound = false;
    mID = g.nextPropertyID;
    g.nextPropertyID++;
    mIDtoProperty.put(new Integer(mID), this);

    mIsOutput = true;
    mIsInput = true;
    mIsEnclosureOut = false;
    mIsEnclosureIn = false;
    mIsFreeFloating = false;
  }

  public boolean isOutput() {return mIsOutput;}
  public boolean isInput() {return mIsInput;}
  public boolean isEnclosureOut() {return mIsEnclosureOut;}
  public boolean isEnclosureIn() {return mIsEnclosureIn;}
  public boolean isFreeFloating() {return mIsFreeFloating;}

  public void setIsOutput(boolean isOutput) {mIsOutput = isOutput;}
  public void setIsInput(boolean isInput) {mIsInput = isInput;}
  public void setIsEnclosureOut(boolean isEnclosureOut) {mIsEnclosureOut = isEnclosureOut;}
  public void setIsEnclosureIn(boolean isEnclosureIn) {mIsEnclosureIn = isEnclosureIn;}
  public void setIsFreeFloating(boolean isFreeFloating) {mIsFreeFloating = isFreeFloating;}

  public static Property newProperty(String name, int type) {
    try {
      Property newProp = (Property)g.propertyTypes[type].mClass.newInstance();
      newProp.mName = name;
      return newProp;
    }
    catch (Exception e) {
      return null;
    }
  }

  public void validate() {
    g.validateCycle++;
    validateLoop();
  }

  private void validateLoop() {
    mGetCycle = g.validateCycle;

    if (mParent != null) {
      if (mParent.mGetCycle != g.validateCycle)
        mParent.validateLoop();
      copyValue(mParent);
      return;
    }

    if (isFresh())
      return;
        // If it's not fresh then it's on output that needs recalculation
    mOwner.update();
  }

  public void copyValue(Property property) {
  }

  public void valueSet() {
    g.setCycle++;
    setLoop();
  }

  private void setLoop() {

    if (mSetCycle == g.setCycle) {
      markOwnerForRefresh();
      return;
    }
    mSetCycle = g.setCycle;
    if (mParent != null && mParent.isSettable()) {
      mParent.copyValue(this);
      mParent.setLoop();
      return;
    }

    markOwnerForRefresh();
  }

    public boolean isSettable() {
        if (mInIsSettable)
            return true;
        mInIsSettable = true;
        if (mParent == null) {
            mInIsSettable = false;
            return mIsInput;
        }
        tempBoolean = mParent.isSettable();
        mInIsSettable = false;
        return tempBoolean;
    }

  public boolean isCompatable(Property property) {
    return (g.propertyTypes[mType].isCompatable(property.getType()) &&
        isArray == property.isArray);
  }

  public boolean isFresh() {
    if (mIsInput) {
      if (mParent == null)    // Unconstrained inputs
        return true;
      return mParent.isFresh();  // it's an output only:
    }
    return mOwner.mFresh;
  }

  void delete() {
    if (mOwner != null && !mOwner.isFunction() && !mOwner.isEnclosure()) {  // Don't delete properties from a graphic entity
      return;
    }

    clearConnections();
    if (mOwner != null) {
      mOwner.removeProperty(this);
    }
  }

  public void setParent(Property parent) {
    mParent = parent;
  }

  public void addChild(Property property) {
    if (Util.contains(mChildren, property))
      return;
    mChildren = (Property[])Util.append(mChildren, Types.PROPERTY, property);
//    Arrays.sort(mChildren);
  }

  public void removeChild(Property property) {
    if (! Util.contains(mChildren, property))
      return;
    mChildren = (Property[])Util.remove(mChildren, Types.PROPERTY, property);
  }

  public void clearConnections() {
    if (mParent != null) {
      mParent.removeChild(this);
      mParent = null;
    }

    for (int i = 0; i < mChildren.length; i++) {
      mChildren[i].setParent(null);
    }
    mChildren = new Property[0];
  }

  public static boolean newRelationship(Property parent, Property child) {
    if (child == parent)
      return false;
    if (!parent.getType().isCompatable(child.getType()))
      return false;
    if (child.mParent == parent)
      return false;                   // Relationship already exists
    if (parent != null && child != null) {
      if (child.mParent != null) {
        child.mParent.removeChild(child);
      }

      child.copyValue(parent);

      parent.addChild(child);
      child.setParent(parent);
      child.valueSet();
    }

    g.scene.refresh();
    return true;
  }

  public PropertyType getType() {
    return g.propertyTypes[mType];
  }

  public void markChildrenForRefresh() {
    for (int i = 0; i < mChildren.length; i++) {
        if (mChildren[i].mOwner.isEnclosure() && !mChildren[i].mOwner.equals(mOwner))
            mChildren[i].markChildrenForRefresh();
        else
            mChildren[i].markOwnerForRefresh();
    }
  }

  public void markOwnerForRefresh() {
    if (mOwner == null)
      return;

    if (mIsInput) {
      mOwner.needsRefresh();
    }
    else if (!mOwner.isGraphicEntity()) {
      markChildrenForRefresh();
    }
    else if (mIsOutput && !mIsInput) {
      markChildrenForRefresh();
    }
  }

  public String valueAsString() {
    validate();
    return "";
  }

  public void insertAt(int index) {
  }

  public void deleteItem(int index) {
  }

  public String getValueByName(String fieldName) {  // used for compound types
    return "";
  }

  public String getValueByName(String fieldName, int index) {  // used for compound types
    return "";
  }

  public void setValueByName(String fieldName, String value) {  // used for compound types
  }

  public void setValueByName(String fieldName, int index, String value) {  // used for compound types
  }

  public int arrayLength() {
    return 0;
  }

  public void setArrayLength(int length) {
  }

  public boolean serializeProperty(XMLWriter xmlOut) {
    try {
      Util.serializeElement(xmlOut, "type", mType);
      Util.serializeElement(xmlOut, "name", mName);
      Util.serializeElement(xmlOut, "ID", mID - g.fileWritingPropertyIDOffset);
      Util.serializeElement(xmlOut, "isInput", mIsInput);
      Util.serializeElement(xmlOut, "isOutput", mIsOutput);
      Util.serializeElement(xmlOut, "isEnclosureIn", mIsEnclosureIn);
      Util.serializeElement(xmlOut, "isEnclosureOut", mIsEnclosureOut);
      Util.serializeElement(xmlOut, "isFreeFloating", mIsFreeFloating);
      if (mIsFreeFloating) {
        Util.serializeElement(xmlOut, "freeFloatingLocationX", g.sceneGraphPanel.getInputNode(this).mFreeFloatingLocation.getX());
        Util.serializeElement(xmlOut, "freeFloatingLocationY", g.sceneGraphPanel.getInputNode(this).mFreeFloatingLocation.getY());
      }
      Util.serializeElement(xmlOut, "isArray", isArray);
      Util.serializeElement(xmlOut, "isCompound", isCompound);
      if (mParent != null)
        if (! (mOwner.isEnclosure() &&
               ((Enclosure)mOwner).mIsTopLevelStored &&
               isEnclosureIn()))   // Don't store parents of inputs to the top level enclosure
          Util.serializeElement(xmlOut, "parent", mParent.mID - g.fileWritingPropertyIDOffset);
    }
    catch (Exception e) {
      return false;
    }
    return true;
  }
}