package roebling;

import javax.swing.JPanel;

/**
 * Title:        ParaMour
 * Description:  Parametric design tool.
 * Copyright:    Copyright (c) 2001
 * Company:      NewBricks
 * @author Simon Greenwold
 * @version 1.0
 */

import java.awt.geom.*;
import java.awt.event.*;
import javax.swing.*;
import java.awt.*;

public class PropertyNode extends PropertyBound {

    protected static final int DIAMETER = 10;
    protected Ellipse2D mBounds;
    protected Ellipse2D mArrayBounds;

    private Color mColor = null;
    public static final Color mNumberColor = Color.red;
    public static final Color mPairColor = Color.blue;
    public static final Color mStringColor = Color.green;
    public static final Color mBooleanColor = Color.yellow;
    public static final Color mColorColor = Color.cyan;
    public static final Color mCustomColor = Color.white;
    public static final Color mDefaultColor = Color.yellow;
    private BasicStroke mBoundsStroke = new BasicStroke(1.0f,
                                                  BasicStroke.CAP_BUTT,
                                                  BasicStroke.JOIN_MITER);
    private BasicStroke mSelectedStroke = new BasicStroke(2.0f,
                                                  BasicStroke.CAP_BUTT,
                                                  BasicStroke.JOIN_MITER);
    public JLabelNoAntialias mPropertyLabel;
    private EnclosureNode mBandEnclosure;

    public int mYLocationOffset;       // The offset into the list of property Y-positions that the entity maintains
    public Point2D mFreeFloatingLocation;  // Location, only used for free-floating property nodes

    private PropertyNode mSelf;

    public PropertyNode() {
    }

    public PropertyNode(Property property, EntityNode entityNode, JLabelNoAntialias propertyLabel) {
        super();
        mProperty = property;
        mEntityNode = entityNode;
        mPropertyLabel = propertyLabel;

        setName(property.mName);
        mSelf = this;
        try {
            jbInit();
        }
        catch(Exception e) {
            e.printStackTrace();
        }
        setOpaque(false);

        addMouseListener(new MouseAdapter() {
            public void mousePressed(MouseEvent e) {
                mBandEnclosure = mEntityNode.getEnclosureNode();
                if ((mBandEnclosure == null && mEntityNode.isEnclosureNode()) ||   // Scene
                    (mEntityNode.isEnclosureNode() && ((EnclosureNode)mEntityNode).isMaximized())) // Maximized
                    mBandEnclosure = (EnclosureNode)mEntityNode;
                Point2D startPoint = new Point2D.Double();
                startPoint.setLocation(mSelf.getCenter(mBandEnclosure.getInterior()));
                mBandEnclosure.startConnectionBand((int)startPoint.getX(), (int)startPoint.getY());

                g.sceneGraphPanel.setActivePropertyNode(mSelf);
            }
            public void mouseReleased(MouseEvent e) {
                mBandEnclosure.stopConnectionBand();
                g.sceneGraphPanel.propertyNodeDragReleased(e);
            }
            public void mouseClicked(MouseEvent e) {
              if (e.getClickCount() == 2) {
                mEntityNode.editPropNode(mSelf);
              }
            }

        });
        addMouseMotionListener(new MouseMotionAdapter() {
            public void mouseDragged(MouseEvent e) {
                Point2D dragPoint = new Point2D.Double();
                dragPoint.setLocation(e.getX(), e.getY());
                Util.changeCoordinates(dragPoint, mSelf, mBandEnclosure.getInterior());
                mBandEnclosure.moveConnectionBand((int)dragPoint.getX(), (int)dragPoint.getY());
            }
        });
    }

    public void nameChanged() {
      PropertyNode node;

      node = g.sceneGraphPanel.getInputNode(mProperty);
      if (node != null) {
        if (node.mPropertyLabel != null)
          node.mPropertyLabel.setText(mProperty.mName);
      }
      node = g.sceneGraphPanel.getOutputNode(mProperty);
      if (node != null) {
        if (node.mPropertyLabel != null)
          node.mPropertyLabel.setText(mProperty.mName);
      }
    }

    public void positionLabel() {
        if (mPropertyLabel == null) {
            return;
        }
        if (mProperty.isFreeFloating()) {
          mPropertyLabel.setBounds((int)(getLocation().getX() + getWidth() + 4),
                                   (int)getLocation().getY(),
                                   (int)(mEntityNode.getWidth() - (getLocation().getX() + getWidth() + 10)),
                                   12);
        }
        else {
          mPropertyLabel.setBounds(EntityNode.PROPERTY_LABEL_X_OFFSET, getY() + EntityNode.PROPERTY_LABEL_Y_OFFSET,
                               mEntityNode.getWidth() - (DIAMETER + EntityNode.X_INSET + EntityNode.PROPERTY_LABEL_X_OFFSET), 12);
        }
    }



    public JLabel getLabel() {
        return mPropertyLabel;
    }

    public void setName(String name) {
        if (mPropertyLabel != null)
            mPropertyLabel.setText(name);
    }

    public Color getPropertyNodeColor() {
        if (mColor != null)
            return mColor;

        return g.propertyTypes[mProperty.mType].mColor;
    }

    public boolean contains(int x, int y) {
      if (mBounds == null) setUpGraphics();

      if (!mProperty.isArray)
        return mBounds.contains(x, y);
      return mBounds.contains(x, y) || mArrayBounds.contains(x, y);
    }

    public void setUpGraphics() {
      if (!mProperty.isArray) {
        mBounds = new Ellipse2D.Double(0, 0, DIAMETER, DIAMETER);
        return;
      }

      if (isOnLeft()) {
        mBounds = new Ellipse2D.Double(4, 0, DIAMETER, DIAMETER);
        mArrayBounds = new Ellipse2D.Double(0, 0, DIAMETER, DIAMETER);
      }
      else {
        mBounds = new Ellipse2D.Double(1, 0, DIAMETER, DIAMETER);
        mArrayBounds = new Ellipse2D.Double(5, 0, DIAMETER, DIAMETER);
      }
    }

    public boolean contains(java.awt.Point p) {
        return contains((int)p.getX(), (int)p.getY());
    }

    // Gets center with respect to an enclosure interior
    public java.awt.Point getCenter(Container container) {
        Point2D center = new Point2D.Double(DIAMETER / 2, DIAMETER / 2);

        Util.changeCoordinates(center, this, container);
        return new java.awt.Point((int)center.getX(), (int)center.getY());
    }

    public void paintComponent(Graphics graphics) {
        super.paintComponent(graphics);
        if (mBounds == null) setUpGraphics();

        Graphics2D g2 = (Graphics2D)graphics;
        g2.setPaint(getPropertyNodeColor());
        g2.fill(mBounds);

        if (g.sceneGraphPanel.getActivePropertyNode() == this) {
            g2.setStroke(mSelectedStroke);
            g2.setPaint(Color.orange);
        }
        else {
            g2.setStroke(mBoundsStroke);
            g2.setPaint(Color.black);
        }
        g2.draw(mBounds);

        if (mProperty.isArray) {
          g2.setPaint(getPropertyNodeColor());
          g2.fill(mArrayBounds);

          if (g.sceneGraphPanel.getActivePropertyNode() == this) {
              g2.setStroke(mSelectedStroke);
              g2.setPaint(Color.orange);
          }
          else {
              g2.setStroke(mBoundsStroke);
              g2.setPaint(Color.black);
          }
          g2.draw(mArrayBounds);
        }
    }

    private void jbInit() throws Exception {
        //this.setBackground(Color.green);
        if (!mProperty.isArray)
          setSize(DIAMETER + 1, DIAMETER + 1);
        else
          setSize(DIAMETER + 6, DIAMETER + 1);
    }

    public void setYLocationOffset(int yLocationOffset) {
        mYLocationOffset = yLocationOffset;
    }

    public int getYLocationOffset() {
        return mYLocationOffset;
    }

    public boolean isOnRight() {
        if (isEnclosureIn())
            return false;
        if (isOutput())
            return true;
        return false;
    }

    public boolean isOnLeft() {
        if (isEnclosureOut())
            return false;
        if (isInput())
            return true;
        return false;
    }

    public boolean isOnEnclosureNode(EnclosureNode enclosureNode) {
        if (! getEntity().isEnclosure())
            return false;
        return enclosureNode == (EnclosureNode)getEntityNode();
    }

    public void delete() {
      if (!mEntityNode.isEnclosureNode() && !mEntityNode.isFunctionNode())
        return;             // Don't delete from graphic entities
      clearConnections();
      mProperty.delete();
      mEntityNode.removeProperty(mProperty);
    }

    public int getOwnerLayer(EnclosureNode enclosureNode) {
        return enclosureNode.getLayer(mEntityNode);
    }
  public void setLocation(int x, int y) {
    super.setLocation( x,  y);
    if (mProperty.isFreeFloating()) {
      moveConnections();
      positionLabel();
    }
  }

}