package roebling;

/**
 * Title:        Roebling
 * Description:  Parametric dataflow programming.
 * Copyright:    Copyright (c) 2001
 * Company:
 * @author Simon Greenwold
 * @version 1.0
 */
import java.awt.*;
import java.awt.font.*;
import java.awt.geom.*;
import java.awt.image.*;

public class Text extends GraphicEntity {

  public static int DEFAULT_SIZE = 12;
  public static String DEFAULT_TEXT = "Text";
  private static Font mDefaultFont = null;

  private NumProp mSizeProperty;
  private float mSize;
  private double mStoredSize;
  private LocationProp mLocationProperty;
  private Point2D.Double mLocation = new Point2D.Double();
  private Point2D.Double mPanelLocation = new Point2D.Double();
  private BoolProp mVisibleProperty;
  private BoolProp mSelectableProperty;
  private ColorProp mColorProperty;
  private Color mColor;
  private Color mStoredColor;
  private StringProp mTextProperty;
  private String mText;
  private String mStoredText;

  private Point2D mPreDragLocation = new Point2D.Double();
  private Point2D mPostDragLocation = new Point2D.Double();

  private boolean mNewBuffer = true;
  private int mMaxHeight = 50;
  private Rectangle mArea;
  private Rectangle2D mBounds = new Rectangle2D.Double();
  BufferedImage mBi;
      Graphics2D mBig;
  private boolean mTextGraphicChanged = true;
  private TextLayout mLayout;
  private Font mFont;


  public Text() {
    super();
    mEntityType = "text";
    addProperties();
    if (mDefaultFont == null)
        initDefaultFont();
    mFont = mDefaultFont;
  }

  public Text(String name, Enclosure enclosure) {
    super(name, enclosure);
    mEntityType = "text";
    addProperties();
    if (mDefaultFont == null)
        initDefaultFont();
    mFont = mDefaultFont;
  }


  private static void initDefaultFont() {
    mDefaultFont = new Font("sansserif", Font.PLAIN, DEFAULT_SIZE);
    if (mDefaultFont == null) mDefaultFont = GraphicsEnvironment.getLocalGraphicsEnvironment().getAllFonts()[0];
  }

  protected void addProperties() {
    mSizeProperty  = new NumProp();
    mSizeProperty.mName = "Size";
    mSizeProperty.value = DEFAULT_SIZE;
    mSizeProperty.valueSet();
    addProperty(mSizeProperty);

    mLocationProperty = new LocationProp();
    mLocationProperty.mName = "Location";
    addProperty(mLocationProperty);

    mTextProperty = new StringProp();
    mTextProperty.mName = "Text";
    mTextProperty.value = DEFAULT_TEXT;
    mTextProperty.valueSet();
    addProperty(mTextProperty);

    mColorProperty = new ColorProp();
    mColorProperty.mName = "Color";
    mColorProperty.a = 255;
    addProperty(mColorProperty);

    mVisibleProperty = new BoolProp();
    mVisibleProperty.mName = "Visible";
    mVisibleProperty.value = true;
    addProperty(mVisibleProperty);

    mSelectableProperty = new BoolProp();
    mSelectableProperty.mName = "Selectable";
    mSelectableProperty.value = true;
    addProperty(mSelectableProperty);
  }

  protected void readInputs() {
    mSizeProperty.validate();
    mSize = (int)java.lang.Math.max(mSizeProperty.value, 0.0);

    if (mStoredSize != mSize) {
      mFont = new Font("sansserif", Font.PLAIN, (int)mSize);
    }
    mStoredSize = mSize;

    mColorProperty.validate();
    mColor = mColorProperty.getColorValue();
    mLocationProperty.validate();
    mPanelLocation.setLocation(g.transform.toPanelLocation(new Point2D.Double(mLocationProperty.x, mLocationProperty.y)));
    mVisibleProperty.validate();
    mVisible = mVisibleProperty.value;
    mSelectableProperty.validate();
    mSelectable = mSelectableProperty.value;
    mTextProperty.validate();
    mText = mTextProperty.value;
    if (mText.equals("")) mText = " ";
  }
/*
  private void storeValues() {
      mStoredSize = mSize;
      mStoredText = mText.toString();
      mStoredColor = Util.copyColor(mColor);
  }

  private boolean hasChanged() {
    if (mStoredSize != mSize) {
      mFont = new Font("sansserif", Font.PLAIN, (int)mSize);
      return true;
    }
    if (!mColor.equals(mStoredColor))
      return true;
    if (!mText.equals(mStoredText))
      return true;
    return false;
  }
*/

  public void setText(String text) {
    mTextProperty.value = text;
    mTextProperty.valueSet();
  }

  public void setColor(double r, double g, double b) {
    setColor(r,g,b, 255);
  }
  public void setColor(double r, double g, double b, double a) {
    mColorProperty.r = r;
    mColorProperty.g = g;
    mColorProperty.b = b;
    mColorProperty.a = a;
    mColorProperty.valueSet();
  }

  public void setSize(double size) {
    mSizeProperty.value = size;
    mSizeProperty.valueSet();
  }

  public void setLocation(double x, double y) {
    mLocationProperty.x = x;
    mLocationProperty.y = y;
    mLocationProperty.valueSet();
  }

  public void setLocation(Point2D point) {
    mLocationProperty.x = point.getX();
    mLocationProperty.y = point.getY();
    mLocationProperty.valueSet();
  }

  public boolean hit(Point2D hitLocation) {
    if (!mSelectable)
      return false;
    if (mLayout == null) return false;
    mEnclosure.inverseTransform(hitLocation,mTransHitPoint);
    return mBounds.intersects(mTransHitPoint.getX() - mPanelLocation.getX() - 2, mTransHitPoint.getY() - mPanelLocation.getY() - 2, 4, 4);
  }

  public void moveTo(Point2D newLocation) {
    mLocationProperty.x = g.transform.toGeomLocation(newLocation).getX();
    mLocationProperty.y = g.transform.toGeomLocation(newLocation).getY();
    mLocationProperty.valueSet();
  }

  public void move(Point2D mouseDelta) {
    mPostDragLocation.setLocation(mPreDragLocation.getX() + mouseDelta.getX(),
                                  mPreDragLocation.getY() + mouseDelta.getY());
    moveTo(mPostDragLocation);
  }

  public void savePreDragLocation() {
    mPreDragLocation.setLocation(mPanelLocation);
  }

  public void draw(Graphics2D graphics) {
//    update();
    if (! mVisible)
      return;

    if(mNewBuffer){
      Dimension dim = new Dimension(500,50); //getScene().getScenePanel().getSize();
      int w = dim.width;
      int h = mMaxHeight;
      mArea = new Rectangle(0, 0, w, h);
      mBi = new BufferedImage(w, h, BufferedImage.TYPE_INT_ARGB);

      mBig = mBi.createGraphics();
      mBig.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
      mBig.setBackground(new Color(1.0f, 1.0f, 1.0f, 0.0f));
      mNewBuffer = false;
    }

    double yStart = mSize * 1.2;
      double yPos = yStart;

/*        if (mTextGraphicChanged) {

        mBig.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC,1.0f));
        // Clears the rectangle that was previously drawn.
        mBig.clearRect(0, 0, mArea.width, mArea.height);

        // Draws and fills the newly positioned rectangle to the buffer.
//                if (mIsSelected)
//                    big.setPaint(Governor.getInstance().getSelectionColor());
//                else

        mBig.setPaint(mColor);

        mLayout = new TextLayout(mText, mFont, mBig.getFontRenderContext());
        mLayout.draw(mBig, 0, (int) yPos);
//            if (first) {
            mBounds.setRect(mLayout.getBounds());
//                first = false;
//            }
//            else {
//                mBounds.setRect(mBounds.getX(), mBounds.getY(), java.lang.Math.max(mBounds.getWidth(), mLayout.getBounds().getWidth()), _bounds.getHeight() + _layout.getAscent() + _layout.getDescent());
//            }

//                    yPos += mLayout.getAscent() + mLayout.getDescent();

        mTextGraphicChanged = false;
    }

    // Draws the buffered image to the screen.

    g.drawImage(mBi, (int)mPanelLocation.getX(), (int)(mPanelLocation.getY() - yStart), null);//_scene.getScenePanel());
*/
    mEnclosure.makeTransform();
    mSaveTrans = graphics.getTransform();
    graphics.transform(mEnclosure.mCompositeTrans);

    graphics.setFont(mFont);
    graphics.setColor(mColor);
    //g.drawString(mText, (int)mPanelLocation.getX(), (int)mPanelLocation.getY());
    mLayout = new TextLayout(mText, mFont,graphics.getFontRenderContext());
    mLayout.draw(graphics, (int)mPanelLocation.getX(), (int)mPanelLocation.getY());
    mBounds.setRect(mLayout.getBounds());

    graphics.setTransform(mSaveTrans);
  }

  public static Font getDefaultFont() {
    if (mDefaultFont == null) initDefaultFont();
    return mDefaultFont;
  }
/*
  private static void initDefaultFont() {
      mDefaultFont = new Font("sansserif", Font.PLAIN, DEFAULT_SIZE);
      if (mDefaultFont == null) mDefaultFont = GraphicsEnvironment.getLocalGraphicsEnvironment().getAllFonts()[0];
  }*/

}