package roebling;

/**
 * Title:        Roebling
 * Description:  Parametric dataflow programming.
 * Copyright:    Copyright (c) 2001
 * Company:
 * @author Simon Greenwold
 * @version 1.0
 */
import java.awt.*;
import javax.swing.*;
import java.awt.geom.*;
import java.awt.event.*;

public class TransformationNode extends EntityNode {

  private static final Color TRANFORMATION_NODE_COLOR = new Color(91, 121, 204, 150);
  private Transformation mTransformation;
  private TransConnector mConnector = null;
  private JButton mUpButton;
  private JButton mDownButton;
  private TransformationNode mNextNode;
  private TransformationNode mPreviousNode;
  private TransformationNode mSelf;

  public TransformationNode() {
    super();
    mSelf = this;
    addButtons();
  }

  protected void loadEntityImage() {
    if (mEntity == null)
      return;
    else if (((Transformation)mEntity).mEntityType.equals("rotation")) {
      mEntityImage = new ImageIcon("images/smallRotation.gif");
    }
    else if (((Transformation)mEntity).mEntityType.equals("translation")) {
      mEntityImage = new ImageIcon("images/smallTranslation.gif");
    }
    else if (((Transformation)mEntity).mEntityType.equals("scale")) {
      mEntityImage = new ImageIcon("images/smallScale.gif");
    }
  }

  public boolean isTransformationNode() {
    return true;
  }

  public void setEntity(Entity entity) {
    super.setEntity(entity);
    mTransformation = (Transformation)mEntity;
  }

  public void makeConnector() {
    if (mConnector != null) {
      mEnclosureNode.getInterior().remove(mConnector);
    }
    if (mNextNode == null)
      return;

    mConnector = new TransConnector();
    mConnector.mFromNode = this;
    mConnector.mToNode = mNextNode;
    mConnector.moveSelf();
    mEnclosureNode.getInterior().add(mConnector);
  }

  private void connectFromPoint(Point2D from) {
    from.setLocation(getLocation().getX() + (getWidth() / 2.0), getLocation().getY() + getHeight());
  }

  private void connectToPoint(Point2D to) {
    to.setLocation(getLocation().getX() + (getWidth() / 2.0), getLocation().getY());
  }

  private TransformationNode nextTransNode() {
    Transformation[] transes = mEntity.mEnclosure.mTransforms;
    int index = Util.find(transes, mTransformation);
    if (index < 0 || index == transes.length - 1)
      return null;
    return (TransformationNode)g.sceneGraphPanel.getEntityNode(transes[index+1]);
  }

  private TransformationNode previousTransNode() {
    Transformation[] transes = mEntity.mEnclosure.mTransforms;
    int index = Util.find(transes, mTransformation);
    if (index < 0 || index == 0)
      return null;
    return (TransformationNode)g.sceneGraphPanel.getEntityNode(transes[index-1]);
  }


  public void resized() {
    super.resized();
    mUpButton.setBounds(20, getHeight() - 17, 16, 12);
    mDownButton.setBounds(36, getHeight() - 17, 16, 12);
    moveAllAttachedConnectors();
  }

  public void transformationListChanged() {
    mNextNode = nextTransNode();
    mPreviousNode = previousTransNode();

    if (mNextNode == null)
      mDownButton.setVisible(false);
    else
      mDownButton.setVisible(true);

    if (mPreviousNode == null)
      mUpButton.setVisible(false);
    else
      mUpButton.setVisible(true);
  }

  private void moveAllAttachedConnectors() {
    if (mPreviousNode != null)
      mPreviousNode.moveConnector();
    moveConnector();
  }

  public void setLocation(java.awt.Point p) {
    super.setLocation(p);
    moveAllAttachedConnectors();
  }

  public void setLocation(int x, int y) {
    super.setLocation(x, y);
    moveAllAttachedConnectors();
  }

  private void moveConnector() {
    if (mConnector == null) return;
    mConnector.moveSelf();
  }

  public Color getFillColor() {
    return TRANFORMATION_NODE_COLOR;
  }

  public void delete() {
    super.delete();
    if (mConnector != null)
      mEnclosureNode.getInterior().remove(mConnector);
    mEnclosureNode.transformationListChanged();
  }

  private void addButtons() {
    mUpButton = new JButton(new ImageIcon("images/up.gif"));
    mUpButton.setSelectedIcon(new ImageIcon("images/upHilite.gif"));

    mUpButton.addActionListener(new ActionListener()  {
      public void actionPerformed(ActionEvent e) {
        mEntity.mEnclosure.moveTransformationUp(mTransformation);
        java.awt.Point tempPoint = getLocation();
        mSelf.setLocation(mPreviousNode.getLocation());
        mPreviousNode.setLocation(tempPoint);
        mEnclosureNode.transformationListChanged();
      }
    });
    add(mUpButton);

    mDownButton = new JButton(new ImageIcon("images/down.gif"));
    mDownButton.setSelectedIcon(new ImageIcon("images/downHilite.gif"));
    mDownButton.addActionListener(new ActionListener()  {
      public void actionPerformed(ActionEvent e) {
        mEntity.mEnclosure.moveTransformationDown(mTransformation);
        java.awt.Point tempPoint = getLocation();
        mSelf.setLocation(mNextNode.getLocation());
        mNextNode.setLocation(tempPoint);
        mEnclosureNode.transformationListChanged();
      }
    });
    add(mDownButton);
  }


  public class TransConnector extends JPanel {
    public TransformationNode mFromNode;
    public TransformationNode mToNode;
    private Point2D mFromPoint = new Point2D.Double();
    private Point2D mToPoint = new Point2D.Double();
    private int top, left, bottom, right;
    private Stroke mConStroke = new BasicStroke(3);
    private Color mColor = new Color(0, 255, 0);

    TransConnector() {
      setOpaque(false);
      setBounds(0, 0, 10, 10);
    }

    private void moveSelf() {
      establishBounds();
      setBounds(left, top, right - left, bottom - top);
    }

    private void establishBounds() {
      mFromNode.connectFromPoint(mFromPoint);
      mToNode.connectToPoint(mToPoint);

      top = (int)Math.min(mFromPoint.getY(), mToPoint.getY()) - 3;
      bottom = (int)Math.max(mFromPoint.getY(), mToPoint.getY()) + 3;
      left = (int)Math.min(mFromPoint.getX(), mToPoint.getX()) - 3;
      right = (int)Math.max(mFromPoint.getX(), mToPoint.getX()) + 3;
    }

    public void paintComponent(Graphics graphics) {
      Graphics2D g2 = (Graphics2D)graphics;
      g2.setStroke(mConStroke);
      g2.setPaint(mColor);
      g2.drawLine((int)mFromPoint.getX() - left,
                  (int)mFromPoint.getY() - top,
                  (int)mToPoint.getX() - left,
                  (int)mToPoint.getY() - top);
    }

  }

}